unit uNXCComp;

interface

uses
  Classes;

type
  TSymbolType = (stUnknown, stParam, stLocal, stGlobal, stAPIFunc, stAPIStrFunc);
  TOnCompilerMessage = procedure(const msg : string; var stop : boolean) of object;
  TNXCComp = class
  private
    fNBCSrc : TStrings;
    fMessages : TStrings;
    fMS : TMemoryStream;
    fTempChar : Char;
    fCCSet : boolean;
    fIncludeDirs: TStrings;
    fCurFile: string;
    fOnCompMSg: TOnCompilerMessage;
    fDirLine : string;
    fCurrentLine : string;
    fAPIFunctions : TStrings;
    fAPIStrFunctions : TStrings;
    fThreadNames : TStrings;
    fCurrentThreadName : string;
    fBytesRead : integer;
    fParamNames : TStrings;
    fSwitchFixups : TStrings;
    fInSwitch : boolean;
    fSwitchIsString : boolean;
    procedure GetCharX;
    procedure GetChar;
    procedure Init;
    procedure Prog;
    procedure SkipCommentBlock;
    procedure SkipLine;
    procedure SkipDirectiveLine;
    procedure SkipWhite;
    procedure GetName;
    procedure GetNum;
    procedure GetHexNum;
    procedure GetCharLit;
    procedure GetOp;
    procedure Next;
    procedure MatchString(x: string);
    procedure Semi;
    procedure Factor;
    procedure Modulo;
    procedure Divide;
    procedure Multiply;
    procedure Term;
    procedure Add;
    procedure Equal;
    procedure EqualString;
    procedure LessString;
    procedure Expression;
    procedure LessOrEqual;
    procedure NotEqual;
    procedure Subtract;
    procedure BoolExpression;
    procedure Greater;
    procedure LeftShift;
    procedure Less;
    procedure NEqual;
    procedure Relation;
    procedure StoreZeroFlag;
    function ValueIsStringType : boolean;
    procedure RightShift;
    procedure BoolOr;
    procedure BoolTerm;
    procedure BoolXor;
    procedure NotFactor;
//    procedure StringConst;
    function  GetParamName(procname : string; idx : integer) : string;
    procedure DoCall(procname: string);
    procedure DoCallAPIFunc(procname: string);
    function  APIFuncNameToID(procname : string) : integer;
    function  IsAPIFunc(procname : string) : boolean;
    procedure Assignment;
    procedure Block(const lend : string = ''; const lstart : string = '');
    procedure BlockStatements(const lend : string = ''; const lstart : string = '');
    procedure CheckBytesRead(const oldBytesRead : integer);
    procedure DoFor;
    procedure DoIf(const lend, lstart : string);
    procedure DoWhile;
    procedure DoDoWhile;
    procedure DoRepeat;
    procedure DoAsm(var dt : char);
    function DecorateVariables(const asmStr : string) : string;
    procedure DoSwitch;
    procedure DoSwitchCase;
    function  GetCaseConstant : string;
    procedure DoSwitchDefault;
    function SwitchFixupIndex : integer;
    procedure FixupSwitch(idx : integer; lbl : string);
    procedure DoLabel;
    procedure Statement(const lend, lstart : string);
    procedure ProcessDirectives;
    procedure HandlePoundLine;
    function ArrayOfType(dt : char) : char;
    function GetVariableType(vt: char; bUnsigned: boolean): char;
    procedure TopDecls;
    procedure IncLineNumber;
    procedure AddLocal(name: string; dt: char);
    procedure AllocGlobal(dt: char);
    procedure AllocLocal(const sub: string; dt: char);
    function GetInitialValue(dt : char) : string;
    procedure DoLocals(const sub: string);
    procedure AddParamName(procname, varname : string; idx : integer);
    function FormalList(protoexists: boolean; procname: string): integer;
    procedure ProcedureBlock;
    procedure FunctionBlock(const Name : string; dt : char);
//    procedure Error(s: string);
    procedure AbortMsg(s: string);
    procedure Expected(s: string);
    procedure Undefined(n: string);
    procedure CheckIdent;
    procedure Duplicate(n: string);
//    function SizeOfType(dt: char): integer;
    procedure AddEntry(N: string; T: char);
    procedure CheckDup(N: string);
    procedure CheckTable(N: string);
    procedure AddParam(N: string; T: char);
    function DataType(const n: string): char;
    procedure LoadVar(const Name: string);
    procedure CheckNotProc(const Name : string);
    procedure Store(const Name: string);
    procedure Allocate(const Name, aVal, Val: string; dt: char);
    procedure Epilog(bIsSub: boolean);
    procedure Header;
    procedure Prolog(const name: string; bIsSub: boolean);
    procedure Trailer;
    procedure EmitRegisters;
    procedure EmitStackVariables;
//    procedure EmitTypeDefs;
    procedure EmitLn(s: string);
    procedure EmitLnNoTab(s: string);
    procedure PostLabel(L: string);
    procedure LoadConst(n: string);
//    procedure Negate;
    procedure NotIt;
    procedure PopAdd;
    procedure PopAnd;
    procedure PopCmpEqual;
    procedure PopCmpGreater;
    procedure PopCmpGreaterOrEqual;
    procedure PopCmpLess;
    procedure PopCmpLessOrEqual;
    procedure PopCmpNEqual;
    procedure PopMod;
    procedure PopDiv;
    procedure PopLeftShift;
    procedure PopMul;
    procedure PopOr;
    procedure PopRightShift;
    procedure PopSub;
    procedure PopXor;
    procedure PushPrim;
    procedure SetZeroCC;
    procedure Branch(L: string);
    procedure BranchFalse(L: string);
    procedure ClearReg;
    procedure ArrayAssignment(const name : string; dt : char);
    procedure MathAssignment(const name : string);
    procedure StoreAdd(const name: string);
    procedure StoreDiv(const name: string);
    procedure StoreMod(const name: string);
    procedure StoreAnd(const name: string);
    procedure StoreOr(const name: string);
    procedure StoreXor(const name: string);
    procedure StoreAbs(const name: string);
    procedure StoreSign(const name: string);
    procedure StoreShift(bRight : boolean; const name: string);
    procedure StoreMul(const name: string);
    procedure StoreSub(const name: string);
    procedure StoreInc(const name: string; const val: integer = 1);
    procedure StoreDec(const name: string; const val: integer = 1);
    procedure DoWait;
    procedure DoAPICommands(const lend, lstart : string);
    procedure DoResetScreen;
    procedure DoReadButton(idx : integer);
    procedure DoBreakContinue(idx : integer; const lbl : string);
    procedure DoOnFwdRev;
    procedure DoOnFwdRevReg;
    procedure DoOnFwdRevSync;
    procedure DoRotateMotors(idx : integer);
    procedure DoSetSensorTypeMode(idx : integer);
    procedure DoClearSetResetSensor;
    procedure DoTextNumOut(idx : integer);
    procedure DoDrawPoint;
    procedure DoDrawLineRect(idx : integer);
    procedure DoDrawCircle;
    procedure DoDrawGraphic;
    procedure DoPlayToneEx;
    procedure DoPlayFileEx;
    procedure DoAcquireRelease;
    procedure DoExitTo;
    procedure DoSetInputOutput(idx : integer);
    procedure DoStop;
    procedure DoGoto;
    procedure DoPrecedesFollows;
    procedure DoReturn;
    procedure DoStopMotors;
    procedure DoStrCat;
    procedure DoSubString;
    procedure DoStrReplace;
    procedure DoStrToNum;
    procedure ReportProblem(const lineNo: integer; const fName, 
      msg: string; err: boolean);
    procedure Scan;
    function IsWhite(c: char): boolean;
    function IsRelop(c: char): boolean;
    function IsOrop(c: char): boolean;
    function IsAlpha(c: char): boolean;
    function IsDigit(c: char): boolean;
    function IsHex(c: char): boolean;
    function IsAlNum(c: char): boolean;
    function IsAddop(c: char): boolean;
    function IsMulop(c: char): boolean;
    procedure GetString;
    procedure CheckNumeric;
    procedure CheckString;
    procedure LoadAPIFunctions;
    procedure AddAPIFunction(const name : string; id : integer);
    function WhatIs(const n: string): TSymbolType;
    procedure StringExpression(const Name : string; bAdd : boolean = False);
    procedure StringConcatAssignment(const Name : string);
    procedure StringFunction(const Name : string);
    function TempWordName: string;
    function TempByteName: string;
    function TempLongName : string;
    function RegisterName(name : string = '') : string;
    function ZeroFlag : string;
    function tos: string;
    function StrTmpBufName(name : string = '') : string;
    function StrBufName(name : string = '') : string;
    function StrRetValName(name : string = '') : string;
    procedure StoreString(const Name: string);
    procedure AddAPIStringFunction(const name: string; id: integer);
    function APIStrFuncNameToID(procname: string): integer;
    function IsAPIStrFunc(procname: string): boolean;
    procedure DoStrLen;
    procedure DoStrIndex;
    function ReplaceTokens(const line: string): string;
    procedure EmitAsmLines(s: string);
    procedure EmitPoundLine;
    function InTable(n: string): Boolean;
    function IsLocal(n: string): integer;
    function IsParam(n: string): integer;
//    function IsConst(n: string): integer;
//    function IsKnown(n: string): integer;
    function Locate(n: string): integer;
    procedure AllocateHelper(const Name, aVal, Val: string; dt: char);
    function AlreadyDecorated(n : string) : boolean;
    function GetDecoratedValue: string;
    function GetDecoratedIdent(const val: string): string;
    function ApplyDecoration(const pre, val : string) : string;
    procedure PopCmpHelper(const cc: string);
    procedure CmpStringHelper(const cc: string);
    procedure GreaterString;
    procedure NEqualString;
    procedure BoolSubExpression;
    function NewLabel: string;
  protected
    fTmpAsmLines : TStrings;
    fBadProgram : boolean;
    fProgErrorCount : integer;
    procedure InternalParseStream;
    procedure Clear;
    procedure PreProcess;
    property SwitchFixups : TStrings read fSwitchFixups;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Parse(const aFilename : string); overload;
    procedure Parse(aStream : TStream); overload;
    procedure Parse(aStrings : TStrings); overload;
    property NBCSource : TStrings read fNBCSrc;
    property CompilerMessages : TStrings read fMessages;
    property IncludeDirs : TStrings read fIncludeDirs;
    property CurrentFile : string read fCurFile write fCurFile;
    property OnCompilerMessage : TOnCompilerMessage read fOnCompMSg write fOnCompMsg;
  end;

implementation

uses
  SysUtils, StrUtils, Math, uPreprocess, uNQCLexer, uNBCLexer, uGenLexer,
  mwGenericLex {$IFDEF FAST_MM}, FastStrings{$ENDIF};

resourcestring
  sTaskName = 'Task name';
  sVariableName = 'Variable name';
  sTooManyArgs  = 'Too many arguments';
  sStringReturnValue = 'String return value';
  sInvalidAssignment = 'Invalid assignment';
  sMaxParamCountExceeded = 'Max param count exceeded';

{--------------------------------------------------------------}
{ Constant Declarations }

const
  TAB = ^I;
  CR  = ^M;
  LF  = ^J;

var
  LCount : integer = 0;

const
//  MAXSYMSIZE   = 16;
  MAXGLOBALS   = 1000;
  MAXLOCALS    = 64;
//  MAXSTAGE     = 128000;
//  MAXDEFS      = 5000;
//  MAXLINESIZE  = 80;
  MAXPARAMS    = 16;
//  MAXCONSTS    = 100;
//  MAXCONSTSIZE = 80;
  MAXINCL	     = 9;
//  MAXSTACK     = 100;

{--------------------------------------------------------------}
{ Type Declarations }

type
  SymTab = array[1..MAXGLOBALS] of string;
  TabPtr = ^SymTab;
  ParTypeTab = array[0..MAXPARAMS] of char;

{--------------------------------------------------------------}
{ Variable Declarations }

var
  Look: char = LF;              { Lookahead Character }
//  PrevLook : char;
  Token: char;             { Encoded Token       }
  Value: string;           { Unencoded Token     }

var
  slevel : integer = 1;
  linenumber : array[1..MAXINCL] of integer;	// current source line number
//  sname : array[1..MAXINCL] of string;// current source name
  totallines : integer = 0;
  endofallsource : boolean = False;

var
  GS_Name : SymTab;
  GS_Type : array[1..MAXGLOBALS] of char;
  GS_Size : array[1..MAXGLOBALS] of integer;	// size (in 'data type' units)
  GS_ReturnType : array[1..MAXGLOBALS] of char; // only for procedures
  NumGlobals : integer = 0;

  GS_ParamCount : array[1..MAXGLOBALS] of integer;
  GS_ParamType : array[1..MAXGLOBALS] of ParTypeTab;

  PS_Name : array[1..MAXPARAMS] of string;
  PS_Type : array[1..MAXPARAMS] of char;
  NumParams : integer = 0;

  LS_Name : array[1..MAXLOCALS] of string;
  LS_Type : array[1..MAXLOCALS] of char;
  NumLocals : integer = 0;
//  TotLocSize : integer;

//  CS_Name : array[1..MAXCONSTS] of string;	// constant symbol table (name)
//  CS_Type : array[1..MAXCONSTS] of char;	// data type
//  CS_Value : array[1..MAXCONSTS] of string;
//  CCount : integer = 0;		// # of autogen constants
//  NumConsts : integer = 0;		// total # of constants

{--------------------------------------------------------------}
{ Definition of Keywords and Token Types }

const TOK_IDENTIFIER		= 'x';
const TOK_IF			      = 'i';
const TOK_ELSE		      = 'l';
const TOK_DO            = 'd';
const TOK_ASM           = 'a';
const TOK_REPEAT        = 'r';
const TOK_SWITCH        = 's';
const TOK_DEFAULT       = 'D';
const TOK_CASE          = 'c';
const TOK_WHILE		      = 'w';
const TOK_FOR			      = 'f';
const TOK_END			      = '}';
const TOK_APISTRFUNC    = 'E';
const TOK_APIFUNC       = 'F';
const TOK_PROCEDURE		  = 'R';
const TOK_TASK          = 'K';
const TOK_BEGIN		      = '{';
const TOK_DIRECTIVE		  = '#';
const TOK_API           = 'Z';
const TOK_LABEL         = 'B';

const TOK_NUM			      = 'N';
const TOK_HEX			      = 'H';
const TOK_UNSIGNED		  = 'U';
const TOK_CHARDEF		    = 'C';
const TOK_SHORTDEF 	    = 'I';
const TOK_LONGDEF       = 'L';
const TOK_BYTEDEF       = 'b';
const TOK_USHORTDEF     = #06;
const TOK_ULONGDEF      = #05;
const TOK_MUTEXDEF      = 'M';
const TOK_STRINGDEF     = 'S';
const TOK_STRINGLIT     = 'G';
const TOK_ARRAYCHARDEF   = #07;
const TOK_ARRAYSHORTDEF  = #08;
const TOK_ARRAYLONGDEF   = #09;
const TOK_ARRAYBYTEDEF   = #14;
const TOK_ARRAYUSHORTDEF = #15;
const TOK_ARRAYULONGDEF  = #16;

const TOK_BLOCK_COMMENT = #01;
const TOK_LINE_COMMENT  = #02;

const
  NKW  = 23; //18;
  NKW1 = 24; //19;

const
  KWlist: array[1..NKW] of string =
              ('if', 'else', {'endif', }'while', {'endwhile',}
               'for', {'endfor', }'sub', 'void', 'task',
               'do', 'repeat', 'switch', 'asm',
               'default', 'case',
               'unsigned', 'long', 'short', 'int',
               'char', 'bool', 'byte', 'mutex', 'string',
               'end');

const                                     // 'xileweRWve'
  KWcode: array[1..NKW1+1] of Char =
    (TOK_IDENTIFIER, TOK_IF, TOK_ELSE,
		 {TOK_ENDIF, }TOK_WHILE, {TOK_ENDWHILE,}
     TOK_FOR, {TOK_ENDFOR, }TOK_PROCEDURE, TOK_PROCEDURE, TOK_TASK,
     TOK_DO, TOK_REPEAT, TOK_SWITCH, TOK_ASM,
     TOK_DEFAULT, TOK_CASE,
     TOK_UNSIGNED, TOK_LONGDEF, TOK_SHORTDEF, TOK_SHORTDEF,
     TOK_CHARDEF, TOK_BYTEDEF, TOK_BYTEDEF, TOK_MUTEXDEF, TOK_STRINGDEF,
     TOK_END,
     #0);

const
  API_BREAK    = 0;
  API_CONTINUE = 1;
  API_WAIT     = 2;
  API_ONFWD    = 3;
  API_ONREV    = 4;
  API_ONFWDREG = 5;
  API_ONREVREG = 6;
  API_ONFWDSYNC = 7;
  API_ONREVSYNC = 8;
  API_COAST     = 9;
  API_OFF       = 10;
  API_ROTATEMOTOR = 11;
  API_ROTATEMOTOREX = 12;
  API_SETSENSORTYPE = 13;
  API_SETSENSORMODE = 14;
  API_CLEARSENSOR = 15;
  API_SETSENSORTOUCH = 16;
  API_SETSENSORLIGHT = 17;
  API_SETSENSORSOUND = 18;
  API_SETSENSORLOWSPEED = 19;
  API_RESETSENSOR = 20;
  API_RETURN = 21;
  API_ACQUIRE = 22;
  API_RELEASE = 23;
  API_PRECEDES = 24;
  API_FOLLOWS  = 25;
  API_EXITTO = 26;
  API_SETINPUT = 27;
  API_SETOUTPUT = 28;
  API_STOP = 29;
  API_GOTO = 30;
  API_FLOAT = 31;

  APICount = 32;
  APIList : array[0..APICount-1] of string = (
    'break', 'continue', 'Wait',
    'OnFwd', 'OnRev', 'OnFwdReg', 'OnRevReg', 'OnFwdSync',
    'OnRevSync', 'Coast', 'Off', 'RotateMotor', 'RotateMotorEx',
    'SetSensorType', 'SetSensorMode', 'ClearSensor',
    'SetSensorTouch', 'SetSensorLight', 'SetSensorSound', 'SetSensorLowspeed',
    'ResetSensor',
    'return', 'Acquire', 'Release', 'Precedes', 'Follows',
    'ExitTo', 'SetInput', 'SetOutput', 'Stop', 'goto', 'Float'
  );


function Replace(const str : string; const src, rep : string) : string;
begin
{$IFDEF FAST_MM}
  Result := FastReplace(str, src, rep);
{$ELSE}
  Result := StringReplace(str, src, rep, [rfReplaceAll]);
{$ENDIF}
end;

var
//  outp : TextFile;
  sd : integer;
  maxsd : integer;

procedure pop;
begin
  dec(sd);
end;

procedure push;
begin
  inc(sd);
  maxsd := Max(maxsd, sd);
end;


procedure TNXCComp.GetCharX;
var
  bytesread : integer;
begin
//  PrevLook := Look;
  bytesread := fMS.Read(Look, 1);
  inc(fBytesRead, bytesread);
//  Read(inp, Look);
  fCurrentLine := fCurrentLine + Look;
  if Look = LF then
  begin
    IncLineNumber;
    fCurrentLine := '';
  end;
  if bytesread < 1 then
    endofallsource := True;
  if endofallsource and (slevel > 1) then begin
    // close file pointer
    linenumber[slevel] := 0;
    dec(slevel);
    Look := LF;
    endofallsource := False;
  end;
end;

{--------------------------------------------------------------}
{ Read New Character From Input Stream }

procedure TNXCComp.GetChar;
begin
  if fTempChar <> ' ' then begin
    Look := fTempChar;
    fTempChar := ' ';
  end
  else begin
    GetCharX;
    if Look = '/' then begin
      fMS.Read(fTempChar, 1);
//      Read(inp, tempchar);
      if fTempChar = '*' then begin
        Look := TOK_BLOCK_COMMENT;
        fTempChar := ' ';
      end
      else if fTempChar = '/' then begin
        Look := TOK_LINE_COMMENT;
        fTempChar := ' ';
      end;
    end;
  end;
end;


(*
{--------------------------------------------------------------}
{ Report an Error }

procedure TNXCComp.Error(s: string);
begin
  fMessages.Add(Format('Error: %s.', [s]));
end;
*)

{--------------------------------------------------------------}
{ Report Error and Halt }

procedure TNXCComp.ReportProblem(const lineNo: integer; const fName,
  msg: string; err : boolean);
var
  tmp, tmp1, tmp2, tmp3, tmp4 : string;
  stop : boolean;
begin
  if lineNo = -1 then
  begin
    tmp := msg;
    fMessages.Add(tmp);
  end
  else
  begin
    if err then
      tmp1 := Format('# Error: %s', [msg])
    else
      tmp1 := Format('# Warning: %s', [msg]);
    fMessages.Add(tmp1);
    tmp2 := Format('File "%s" ; line %d', [fName, lineNo]);
    fMessages.Add(tmp2);
    tmp3 := Format('#   %s', [fCurrentLine]);
    fMessages.Add(tmp3);
    tmp4 := '#----------------------------------------------------------';
    fMessages.Add(tmp4);
    tmp := tmp1+#13#10+tmp2+#13#10+tmp3+#13#10+tmp4;
  end;
  fBadProgram := err;
  if err then
    inc(fProgErrorCount);
  stop := false;
  if assigned(fOnCompMsg) then
    fOnCompMsg(tmp, stop);
  if stop then
    Abort;
end;

procedure TNXCComp.AbortMsg(s: string);
begin
  ReportProblem(linenumber[slevel], CurrentFile, s, True);
//  Next; // advance parser
//  Abort;
end;

(*
{--------------------------------------------------------------}
{Return the size in base units of a standard datatype }

function TNXCComp.SizeOfType(dt : char) : integer;
begin
  case dt of
    TOK_CHARDEF, TOK_BYTEDEF : Result := 1;
    TOK_SHORTDEF, TOK_USHORTDEF : Result := 2;
    TOK_LONGDEF, TOK_ULONGDEF : Result := 4;
    TOK_MUTEXDEF : Result := 4;
  else
    Result := 0;
    AbortMsg('SizeOfType() - Unknown Data Type');
  end;
end;
*)

{--------------------------------------------------------------}
{ Report What Was Expected }

procedure TNXCComp.Expected(s: string);
begin
  AbortMsg(s + ' expected');
end;

{--------------------------------------------------------------}
{ Report an Undefined Identifier }

procedure TNXCComp.Undefined(n: string);
begin
  AbortMsg('Undefined Identifier ' + n);
end;


{--------------------------------------------------------------}
{ Report a Duplicate Identifier }

procedure TNXCComp.Duplicate(n: string);
begin
  AbortMsg('Duplicate Identifier ' + n);
end;


{--------------------------------------------------------------}
{ Check to Make Sure the Current Token is an Identifier }

procedure TNXCComp.CheckIdent;
begin
  if Token <> TOK_IDENTIFIER then Expected('Identifier');
end;

{--------------------------------------------------------------}
{ Check to Make Sure the Current Token is a Number }

procedure TNXCComp.CheckNumeric;
begin
  if not (Token in [TOK_NUM, TOK_HEX]) then Expected('Number');
end;

procedure TNXCComp.CheckString;
begin
  if (Token <> TOK_STRINGLIT) and
     (DataType(Value) <> TOK_STRINGDEF) then
    Expected('string constant or variable of type string');
end;



{--------------------------------------------------------------}
{ Recognize an Alpha Character }

function TNXCComp.IsAlpha(c: char): boolean;
begin
//  Result := upcase(c) in ['_', 'A'..'Z'];
  Result := c in ['A'..'Z', 'a'..'z', '_'];
end;

{--------------------------------------------------------------}
{ Recognize a Decimal Digit }

function TNXCComp.IsDigit(c: char): boolean;
begin
  Result := c in ['0'..'9'];
end;

{--------------------------------------------------------------}
{ Recognize a Hex Digit }

function TNXCComp.IsHex(c: char): boolean;
begin
  Result := IsDigit(c) or (c in ['a'..'f', 'A'..'F']);
end;

{--------------------------------------------------------------}
{ Recognize an Alphanumeric }

function TNXCComp.IsAlNum(c: char): boolean;
begin
  Result := IsAlpha(c) or IsDigit(c){ or (c = '_')};
end;

{--------------------------------------------------------------}
{ Recognize an Addop }

function TNXCComp.IsAddop(c: char) : boolean;
begin
  Result := c in ['+', '-'];
end;

{--------------------------------------------------------------}
{ Recognize a Mulop }

function TNXCComp.IsMulop(c: char): boolean;
begin
  Result := c in ['*', '/', '%'];
end;

{--------------------------------------------------------------}
{ Recognize a Boolean Orop }

function TNXCComp.IsOrop(c: char): boolean;
begin
  Result := c in ['|', '^'];
end;

{--------------------------------------------------------------}
{ Recognize a Relop }

function TNXCComp.IsRelop(c: char): boolean;
begin
  Result := c in ['=', '!', '<', '>'];
end;

{--------------------------------------------------------------}
{ Recognize White Space }

function TNXCComp.IsWhite(c: char): boolean;
begin
//  if c = LF then
//    IncLineNumber;
  Result := c in [' ', TAB, CR, LF, TOK_BLOCK_COMMENT, TOK_LINE_COMMENT];
end;

{--------------------------------------------------------------}
{ Skip A Comment Field }

procedure TNXCComp.SkipCommentBlock;
begin
  repeat
    repeat
      GetCharX;
//      if Look = LF then
//        IncLineNumber;
    until (Look = '*') or endofallsource;
    GetCharX;
//    if Look = LF then
//      IncLineNumber;
  until (Look = '/') or endofallsource;
  GetChar;
end;


{--------------------------------------------------------------}
{ Skip A Comment To End Of Line field }

procedure TNXCComp.SkipLine;
begin
  repeat
    GetCharX;
  until (Look = LF) or endofallsource;
//  if Look = LF then
//    IncLineNumber;
  GetChar;
end;

procedure TNXCComp.SkipDirectiveLine;
begin
  fDirLine := '#' + Value + ' ';
  repeat
    GetCharX;
    fDirLine := fDirLine + Look;
  until (Look = LF) or endofallsource;
//  if Look = LF then
//    IncLineNumber;
  GetChar;
//  Next;
end;

{--------------------------------------------------------------}
{ Skip Over Leading White Space }

procedure TNXCComp.SkipWhite;
begin
  while IsWhite(Look) and not endofallsource do begin
    case Look of
      TOK_LINE_COMMENT : SkipLine;
      TOK_BLOCK_COMMENT : SkipCommentBlock;
    else
      GetChar;
    end;
  end;
end;


{--------------------------------------------------------------}
{ Table Lookup }

function Lookup(T: TabPtr; s: string; n: integer): integer;
var
  i: integer;
  found: Boolean;
begin
  found := false;
  i := n;
  while (i > 0) and not found do
     if s = T^[i] then
        found := true
     else
        dec(i);
  Result := i;
end;


{--------------------------------------------------------------}
{ Locate a Symbol in Table }
{ Returns the index of the entry.  Zero if not present. }

function TNXCComp.Locate(n: string): integer;
begin
  Result := Lookup(@GS_Name, n, NumGlobals);
end;


{--------------------------------------------------------------}
{ Look for Symbol in Table }

function TNXCComp.InTable(n: string): Boolean;
begin
  Result := Lookup(@GS_Name, n, NumGlobals) <> 0;
end;


function TNXCComp.AlreadyDecorated(n: string): boolean;
var
  i : integer;
  tmp : string;
begin
  // a variable is considered to be already decorated if it
  // starts with "__" followed by a task name followed by a "_"
  Result := False;
  i := Pos('__', n);
  if i = 1 then
  begin
    System.Delete(n, 1, 2); // remove the '__' at the beginning
    i := Pos('_', n);
    if i > 1 then
    begin
      tmp := Copy(n, 1, i-1);
      i := fThreadNames.IndexOf(tmp);
      Result := i <> -1;
    end;
  end;
end;

{--------------------------------------------------------------}
{ Look for Symbol in Parameter Table }

function TNXCComp.IsParam(n: string): integer;
begin
  if AlreadyDecorated(n) then
    Result := Lookup(@PS_Name, n, NumParams)
  else
    Result := Lookup(@PS_Name, ApplyDecoration(fCurrentThreadName, n), NumParams);
end;


{--------------------------------------------------------------}
{ Look for Symbol in Local Table }

function TNXCComp.IsLocal(n: string): integer;
begin
  if AlreadyDecorated(n) then
    Result := Lookup(@LS_Name, n, NumParams)
  else
    Result := Lookup(@LS_Name, ApplyDecoration(fCurrentThreadName, n), NumLocals);
end;


(*
{--------------------------------------------------------------}
{ Look for Symbol in Constants Table }

function TNXCComp.IsConst(n: string): integer;
begin
  Result := Lookup(@CS_Name, n, NumConsts);
end;
*)


(*
{--------------------------------------------------------------}
{ Test to see if name is already known }

function TNXCComp.IsKnown(n : string) : integer;
begin
  Result := Locate(n) or IsLocal(n) or IsParam(n) or IsConst(n);
end;
*)

{--------------------------------------------------------------}
{ Check to See if an Identifier is in the Symbol Table         }
{ Report an error if it's not. }

procedure TNXCComp.CheckTable(N: string);
begin
  if not InTable(N) then Undefined(N);
end;


{--------------------------------------------------------------}
{ Check the Symbol Table for a Duplicate Identifier }
{ Report an error if identifier is already in table. }


procedure TNXCComp.CheckDup(N: string);
begin
  if InTable(N) then Duplicate(N);
end;


{--------------------------------------------------------------}
{ Add a New Entry to Symbol Table }

procedure TNXCComp.AddEntry(N: string; T: char);
begin
  CheckDup(N);
  if NumGlobals = MAXGLOBALS then AbortMsg('Symbol Table Full');
  Inc(NumGlobals);
  GS_Name[NumGlobals] := N;
  GS_Type[NumGlobals] := T;
end;


{--------------------------------------------------------------}
{ Get an Identifier }

procedure TNXCComp.GetName;
begin
  SkipWhite;
  if Not IsAlpha(Look) then Expected('Identifier');
  Token := TOK_IDENTIFIER;
  Value := '';
  repeat
    Value := Value + Look;
//    Value := Value + UpCase(Look);
    GetChar;
  until not IsAlNum(Look);
end;


{--------------------------------------------------------------}
{ Get a Number }

procedure TNXCComp.GetNum;
var
  savedLook : char;
begin
  SkipWhite;
  if not IsDigit(Look) then Expected('Number');
  savedLook := Look;
  GetChar;
  if Look in ['x', 'X'] then
  begin
    GetHexNum;
  end
  else
  begin
    Token := TOK_NUM;
    Value := savedLook;
    if not IsDigit(Look) then Exit;
    repeat
      Value := Value + Look;
      GetChar;
    until not IsDigit(Look);
  end;
end;


{--------------------------------------------------------------}
{ Get a Hex Number }

procedure TNXCComp.GetHexNum;
begin
  SkipWhite;
  GetChar(); // skip the $ (or 'x')
  if not IsHex(Look) then Expected('Hex number');
  Token := TOK_HEX;
  Value := '0x';
  repeat
    Value := Value + Look;
    GetChar;
  until not IsHex(Look);
end;


{--------------------------------------------------------------}
{ Get a Character Literal }

procedure TNXCComp.GetCharLit;
begin
  SkipWhite;
  GetChar; // skip the '
  Token := TOK_NUM;
  Value := IntToStr(Ord(Look));
  GetChar;
  if Look <> '''' then Expected('Character literal');
  GetChar;
end;

{--------------------------------------------------------------}
{ Get a string Literal }

procedure TNXCComp.GetString;
begin
  SkipWhite;
  GetChar; // skip the "
  Token := TOK_STRINGLIT;
  Value := '''' + Look;
  repeat
    GetCharX;
    if (Look <> LF) and (Look <> '"') then
    begin
      if Look = '''' then
        Value := Value + '"'
      else
        Value := Value + Look;
    end;
  until (Look = '"') or (Look = LF) or endofallsource;
  Value := Value + '''';
  if Look <> '"' then Expected('String literal');
  GetChar;
end;


{--------------------------------------------------------------}
{ Get an Operator }

procedure TNXCComp.GetOp;
begin
  SkipWhite;
  Token := Look;
  Value := Look;
  GetChar;
end;


{--------------------------------------------------------------}
{ Get the Next Input Token }

procedure TNXCComp.Next;
begin
  SkipWhite;
  if Look = '''' then GetCharLit
  else if Look = '"' then GetString
  else if IsAlpha(Look) then GetName
  else if IsDigit(Look) then GetNum
  else if Look = '$' then GetHexNum
  else GetOp;
end;

function IsAPICommand(const name : string) : boolean;
var
  i : integer;
begin
  Result := False;
  for i := Low(APIList) to High(APIList) do
  begin
    if APIList[i] = name then
    begin
      Result := True;
      Break;
    end;
  end;
end;

{--------------------------------------------------------------}
{ Scan the Current Identifier for Keywords }

procedure TNXCComp.Scan;
var
  idx : integer;
begin
  if Token = TOK_IDENTIFIER then
  begin
    idx := Lookup(Addr(KWlist), Value, NKW);
    if idx <> 0 then
      Token := KWcode[idx + 1]
    else
    begin
      // is it an API command?
      if IsAPICommand(Value) then
        Token := TOK_API;
    end;
  end;
end;


{--------------------------------------------------------------}
{ Match a Specific Input String }

procedure TNXCComp.MatchString(x: string);
begin
  if Value <> x then Expected('''' + x + '''');
  Next;
end;


{--------------------------------------------------------------}
{ Match a Semicolon }

procedure TNXCComp.Semi;
begin
//  MatchString(';');
  if Token = ';' then
    Next;
end;


{--------------------------------------------------------------}
{ Output a String with Tab and CRLF }

procedure TNXCComp.EmitLn(s: string);
begin
  fNBCSrc.Add(#9+s);
end;

procedure TNXCComp.EmitPoundLine;
begin
  fNBCSrc.Add('#line ' + IntToStr(linenumber[slevel]) + ' "' + CurrentFile + '"');
end;


{--------------------------------------------------------------}
{ Output a String without Tab with CRLF }

procedure TNXCComp.EmitLnNoTab(s: string);
begin
  fNBCSrc.Add(s);
end;

procedure TNXCComp.EmitAsmLines(s: string);
begin
  if Pos(#10, s) > 0 then
  begin
    fTmpAsmLines.Text := s;
    fNBCSrc.Add(Format('#pragma macro %d', [fTmpAsmLines.Count]));
    fNBCSrc.AddStrings(fTmpAsmLines);
    EmitPoundLine;
  end
  else
    fNBCSrc.Add(s);
end;


{--------------------------------------------------------------}
{ Generate a Unique Label }

function TNXCComp.NewLabel: string;
var
  S: string;
begin
  S := '';
  Str(LCount, S);
  NewLabel := 'L' + S;
  Inc(LCount);
end;


{--------------------------------------------------------------}
{ Post a Label To Output }

procedure TNXCComp.PostLabel(L: string);
begin
  EmitLnNoTab(L+':');
end;


{---------------------------------------------------------------}
{ Initialize Parameter Table to Null }
procedure ClearParams;
var
  i : integer;
begin
  for i := 1 to MAXPARAMS do
  begin
    PS_Name[i] := '';
    PS_Type[i] := #0;
  end;
  NumParams := 0;
end;

{---------------------------------------------------------------}
{ Initialize Locals Table to Null }
procedure ClearLocals;
var
  i : integer;
begin
  for i := 1 to MAXLOCALS do begin
    LS_Name[i] := '';
    LS_Type[i] := #0;
  end;
  NumLocals := 0;
//  TotLocSize := 0;
end;

{--------------------------------------------------------------}
{ Add a Parameter to Table }

procedure TNXCComp.AddParam(N: string; T: char);
begin
  if IsParam(N) <> 0 then Duplicate(N);
  Inc(NumParams);
  PS_Name[NumParams] := N;
  PS_Type[NumParams] := T;
end;

function TNXCComp.WhatIs(const n : string) : TSymbolType;
begin
  if IsParam(n) <> 0 then Result := stParam
  else if IsLocal(n) <> 0 then Result := stLocal
  else if InTable(n) then Result := stGlobal
  else if IsAPIFunc(n) then Result := stAPIFunc
  else if IsAPIStrFunc(n) then Result := stAPIStrFunc
  else Result := stUnknown;
end;

function TNXCComp.DataType(const n : string) : char;
begin
  if (n = 'true') or (n = 'false') then
    Result := TOK_BYTEDEF
  else
  begin
    case WhatIs(n) of
      stParam : Result := PS_Type[IsParam(n)];
      stLocal : Result := LS_Type[IsLocal(n)];
      stGlobal : Result := GS_Type[Locate(n)];
      stAPIFunc : Result := TOK_APIFUNC;
      stAPIStrFunc : Result := TOK_APISTRFUNC;
    else
      Result := #0;
      AbortMsg('Unknown data type');
    end;
  end;
end;

function TNXCComp.ArrayOfType(dt: char): char;
begin
  case dt of
    TOK_CHARDEF : Result := TOK_ARRAYCHARDEF;
    TOK_SHORTDEF : Result := TOK_ARRAYSHORTDEF;
    TOK_LONGDEF : Result := TOK_ARRAYLONGDEF;
    TOK_BYTEDEF : Result := TOK_ARRAYBYTEDEF;
    TOK_USHORTDEF : Result := TOK_ARRAYUSHORTDEF;
    TOK_ULONGDEF : Result := TOK_ARRAYULONGDEF;
  else
    Result := dt;
  end;
end;

{---------------------------------------------------------------}
{ Add Primary or var }

procedure TNXCComp.StoreAdd(const name : string);
begin
  EmitLn(Format('add %0:s, %0:s, %s', [GetDecoratedIdent(name), RegisterName]));
end;

{---------------------------------------------------------------}
{ Subtract Primary from var }

procedure TNXCComp.StoreSub(const name : string);
begin
  EmitLn(Format('sub %0:s, %0:s, %s', [GetDecoratedIdent(name), RegisterName]));
end;

{---------------------------------------------------------------}
{ Multiply Primary with var }

procedure TNXCComp.StoreMul(const name : string);
begin
  EmitLn(Format('mul %0:s, %0:s, %s', [GetDecoratedIdent(name), RegisterName]));
end;

{---------------------------------------------------------------}
{ Divide Primary with var }

procedure TNXCComp.StoreDiv(const name : string);
begin
  EmitLn(Format('div %0:s, %0:s, %s', [GetDecoratedIdent(name), RegisterName]));
end;

{---------------------------------------------------------------}
{ Mod Primary with var }

procedure TNXCComp.StoreMod(const name : string);
begin
  EmitLn(Format('mod %0:s, %0:s, %s', [GetDecoratedIdent(name), RegisterName]));
end;

procedure TNXCComp.StoreAbs(const name: string);
begin
  EmitLn(Format('abs %0:s, %s', [GetDecoratedIdent(name), RegisterName]));
end;

procedure TNXCComp.StoreAnd(const name: string);
begin
  EmitLn(Format('and %0:s, %0:s, %s', [GetDecoratedIdent(name), RegisterName]));
end;

procedure TNXCComp.StoreOr(const name: string);
begin
  EmitLn(Format('or %0:s, %0:s, %s', [GetDecoratedIdent(name), RegisterName]));
end;

procedure TNXCComp.StoreShift(bRight: boolean; const name: string);
begin
  if bRight then
    EmitLn(Format('shr %0:s, %0:s, %s', [GetDecoratedIdent(name), RegisterName]))
  else
    EmitLn(Format('shl %0:s, %0:s, %s', [GetDecoratedIdent(name), RegisterName]));
end;

procedure TNXCComp.StoreSign(const name: string);
begin
  EmitLn(Format('sign %0:s, %s', [GetDecoratedIdent(name), RegisterName]));
end;

procedure TNXCComp.StoreXor(const name: string);
begin
  EmitLn(Format('xor %0:s, %0:s, %s', [GetDecoratedIdent(name), RegisterName]));
end;

{---------------------------------------------------------------}
{ increment var }

procedure TNXCComp.StoreInc(const name : string; const val : integer);
begin
  EmitLn(Format('add %0:s, %0:s, %d', [GetDecoratedIdent(name), val]));
end;

{---------------------------------------------------------------}
{ decrement var }

procedure TNXCComp.StoreDec(const name : string; const val : integer);
begin
  EmitLn(Format('sub %0:s, %0:s, %d', [GetDecoratedIdent(name), val]));
end;

{---------------------------------------------------------------}
{ Clear the Primary Register }

procedure TNXCComp.ClearReg;
begin
  fCCSet := False;
  EmitLn(Format('set %s, 0', [RegisterName]));
end;

(*
{---------------------------------------------------------------}
{ Negate the Primary Register }

procedure TNXCComp.Negate;
begin
  fCCSet := False;
  EmitLn(Format('neg %0:s, %0:s', [RegisterName]));
end;
*)

{---------------------------------------------------------------}
{ Complement the Primary Register }

procedure TNXCComp.NotIt;
begin
  fCCSet := False;
  EmitLn(Format('not %0:s, %0:s', [RegisterName]));
end;


{---------------------------------------------------------------}
{ Load a Constant Value to Primary Register }

procedure TNXCComp.LoadConst(n: string);
var
  cval : int64;
  tmpSrc : string;
begin
  cval := StrToIntDef(n, 0);
  if (cval < Low(smallint)) or (cval > High(word)) then
    tmpSrc := 'mov %s, %s'
  else
    tmpSrc := 'set %s, %s';
  fCCSet := False;
  EmitLn(Format(tmpSrc, [RegisterName, n]));
end;

procedure TNXCComp.CheckNotProc(const Name : string);
begin
  if DataType(Name) in [TOK_PROCEDURE, TOK_TASK] then
    AbortMsg('Can not assign to a task or subroutine');
end;

{---------------------------------------------------------------}
{ Load a Variable to Primary Register }

procedure TNXCComp.LoadVar(const Name: string);
begin
  CheckNotProc(Name);
  fCCSet := False;
  EmitLn(Format('mov %s, %s', [RegisterName, GetDecoratedIdent(Name)]));
end;

{---------------------------------------------------------------}
{ Push Primary onto Stack }

procedure TNXCComp.PushPrim;
begin
  push;
  EmitLn(Format('mov %s, %s', [tos, RegisterName]));
end;

{---------------------------------------------------------------}
{ Add Top of Stack to Primary }

procedure TNXCComp.PopAdd;
begin
  fCCSet := False;
  EmitLn(Format('add %0:s, %1:s, %0:s', [RegisterName, tos]));
  pop;
end;

{---------------------------------------------------------------}
{ Subtract Primary from Top of Stack }

procedure TNXCComp.PopSub;
begin
  fCCSet := False;
  EmitLn(Format('sub %0:s, %1:s, %0:s', [RegisterName, tos]));
  pop;
end;

{---------------------------------------------------------------}
{ Multiply Top of Stack by Primary }

procedure TNXCComp.PopMul;
begin
  fCCSet := False;
  EmitLn(Format('mul %0:s, %1:s, %0:s', [RegisterName, tos]));
  pop;
end;

{---------------------------------------------------------------}
{ Divide Top of Stack by Primary }

procedure TNXCComp.PopDiv;
begin
  fCCSet := False;
  EmitLn(Format('div %0:s, %1:s, %0:s', [RegisterName, tos]));
  pop;
end;

{---------------------------------------------------------------}
{ Modulo Top of Stack by Primary }

procedure TNXCComp.PopMod;
begin
  fCCSet := False;
  EmitLn(Format('mod %0:s, %1:s, %0:s', [RegisterName, tos]));
  pop;
end;

{---------------------------------------------------------------}
{ AND Top of Stack with Primary }

procedure TNXCComp.PopAnd;
begin
  fCCSet := False;
  EmitLn(Format('and %0:s, %1:s, %0:s', [RegisterName, tos]));
  pop;
end;

{---------------------------------------------------------------}
{ OR Top of Stack with Primary }

procedure TNXCComp.PopOr;
begin
  fCCSet := False;
  EmitLn(Format('or %0:s, %1:s, %0:s', [RegisterName, tos]));
  pop;
end;

{---------------------------------------------------------------}
{ XOR Top of Stack with Primary }

procedure TNXCComp.PopXor;
begin
  fCCSet := False;
  EmitLn(Format('xor %0:s, %1:s, %0:s', [RegisterName, tos]));
  pop;
end;


{---------------------------------------------------------------}
{ Left Shift Top of Stack to Primary }

procedure TNXCComp.PopLeftShift;
begin
  fCCSet := False;
  EmitLn(Format('shl %0:s, %1:s, %0:s', [RegisterName, tos]));
  pop;
end;

{---------------------------------------------------------------}
{ Right Shift Top of Stack to Primary }

procedure TNXCComp.PopRightShift;
begin
  fCCSet := False;
  EmitLn(Format('shr %0:s, %1:s, %0:s', [RegisterName, tos]));
  pop;
end;

{---------------------------------------------------------------}
{ Set zero flag based on __D0 value }

procedure TNXCComp.SetZeroCC;
begin
  fCCSet := True;
  EmitLn(Format('tst NEQ, %s, %s',[ZeroFlag, RegisterName]));
end;

function TNXCComp.ValueIsStringType: boolean;
var
  i : integer;
begin
  Result := IsAPIStrFunc(Value);
  if not Result then
  begin
    Result := (Token = TOK_STRINGLIT) or
              ((Token = TOK_IDENTIFIER) and
               (DataType(Value) = TOK_STRINGDEF));
    if not Result then
    begin
      i := Locate(Value);
      if GS_Type[i] = TOK_PROCEDURE then
        Result := GS_ReturnType[i] = TOK_STRINGDEF;
    end;
  end;
end;

procedure TNXCComp.CmpStringHelper(const cc : string);
begin
  fCCSet := True;
  EmitLn(Format('cmp %s, %s, %s, %s',[cc, ZeroFlag, StrTmpBufName, StrBufName]));
end;

procedure TNXCComp.PopCmpHelper(const cc : string);
begin
  fCCSet := True;
  EmitLn(Format('cmp %s, %s, %s, %s',[cc, ZeroFlag, tos, RegisterName]));
  pop;
end;

{---------------------------------------------------------------}
{ Set __zf If Compare was = }

procedure TNXCComp.PopCmpEqual;
begin
  PopCmpHelper('EQ');
end;

{---------------------------------------------------------------}
{ Set __D0 If Compare was != }

procedure TNXCComp.PopCmpNEqual;
begin
  PopCmpHelper('NEQ');
end;

{---------------------------------------------------------------}
{ Set __D0 If Compare was > }

procedure TNXCComp.PopCmpGreater;
begin
  PopCmpHelper('GT');
end;

{---------------------------------------------------------------}
{ Set __D0 If Compare was < }

procedure TNXCComp.PopCmpLess;
begin
  PopCmpHelper('LT');
end;

{---------------------------------------------------------------}
{ Set __D0 If Compare was <= }

procedure TNXCComp.PopCmpLessOrEqual;
begin
  PopCmpHelper('LTEQ');
end;

{---------------------------------------------------------------}
{ Set __D0 If Compare was >= }

procedure TNXCComp.PopCmpGreaterOrEqual;
begin
  PopCmpHelper('GTEQ');
end;

{---------------------------------------------------------------}
{ Store Primary to Variable }

procedure TNXCComp.Store(const Name: string);
begin
  CheckNotProc(Name);
  EmitLn(Format('mov %s, %s',[GetDecoratedIdent(Name), RegisterName]));
end;

procedure TNXCComp.StoreString(const Name : string);
begin
  CheckNotProc(Name);
  EmitLn(Format('mov %s, %s', [GetDecoratedIdent(Name), StrBufName]));
end;

{---------------------------------------------------------------}
{ Branch Unconditional  }

procedure TNXCComp.Branch(L: string);
begin
  EmitLn('jmp ' + L);
end;

{---------------------------------------------------------------}
{ Branch False }

procedure TNXCComp.BranchFalse(L: string);
begin
  // if the condition code has not been set then set it manually
  if not fCCSet then
    SetZeroCC;
  EmitLn(Format('brtst EQ, %s, %s', [L, ZeroFlag]));
end;

{--------------------------------------------------------------}
{ Write Header Info }

procedure TNXCComp.Header;
begin
  // do nothing
end;

{--------------------------------------------------------------}
{ Write Trailer Info }

procedure TNXCComp.Trailer;
var
  tmp : TStrings;
begin
  // handle stack variables
  tmp := TStringList.Create;
  try
    tmp.AddStrings(fNBCSrc);
    fNBCSrc.Clear;
//    EmitTypeDefs;
    EmitLnNoTab('#include "NXTDefs.h"');
    EmitLnNoTab('dseg segment');
    // structures
    EmitLn('__SSMArgs TSetScreenMode');
    EmitLn('__SPTArgs TSoundPlayTone');
    EmitLn('__SPFArgs TSoundPlayFile');
    // mutexes
    EmitLn('__SSMArgsMutex mutex');
    EmitLn('__SPTArgsMutex mutex');
    EmitLn('__SPFArgsMutex mutex');

    EmitRegisters;
    EmitStackVariables;
    EmitLnNoTab('dseg ends');
    fNBCSrc.AddStrings(tmp);
  finally
    tmp.Free;
  end;
end;

{--------------------------------------------------------------}
{ Write the Prolog }

procedure TNXCComp.Prolog(const name : string; bIsSub : boolean);
begin
  if bIsSub then
    EmitLnNoTab('subroutine ' + name)
  else
    EmitLnNoTab('thread ' + name);
end;

{--------------------------------------------------------------}
{ Write the Epilog }

procedure TNXCComp.Epilog(bIsSub : boolean);
begin
  if bIsSub then
  begin
    EmitLn('return');
    EmitLnNoTab('ends');
  end
  else
    EmitLnNoTab('endt');
  EmitLnNoTab('');
end;

{--------------------------------------------------------------}
{ Allocate Storage for a Static variable }

procedure TNXCComp.AllocateHelper(const Name, aVal, Val: string; dt : char);
begin
  case dt of
    TOK_CHARDEF,
    TOK_ARRAYCHARDEF	 : EmitLn(Format('%s sbyte%s %s', [Name, aVal, Val]));
    TOK_SHORTDEF,
    TOK_ARRAYSHORTDEF  : EmitLn(Format('%s sword%s %s', [Name, aVal, Val]));
    TOK_LONGDEF,
    TOK_ARRAYLONGDEF   : EmitLn(Format('%s sdword%s %s', [Name, aVal, Val]));
    TOK_BYTEDEF,
    TOK_ARRAYBYTEDEF   : EmitLn(Format('%s byte%s %s', [Name, aVal, Val]));
    TOK_USHORTDEF,
    TOK_ARRAYUSHORTDEF : EmitLn(Format('%s word%s %s', [Name, aVal, Val]));
    TOK_ULONGDEF,
    TOK_ARRAYULONGDEF  : EmitLn(Format('%s dword%s %s', [Name, aVal, Val]));
    TOK_MUTEXDEF  : EmitLn(Format('%s mutex', [Name]));
    TOK_STRINGDEF : EmitLn(Format('%s byte[] %s', [Name, Val]));
  else
    EmitLn(Format('%s sword %s', [Name, Val]));
  end;
end;

procedure TNXCComp.Allocate(const Name, aVal, Val: string; dt : char);
begin
  EmitLnNoTab('dseg segment');
  AllocateHelper(Name, aVal, Val, dt);
  EmitLnNoTab('dseg ends');
end;

{---------------------------------------------------------------}
{ Parse and Translate a Math Factor }

procedure TNXCComp.Factor;
var
  savedtoken, rdt : char;
  savedvalue : string;
  idx : integer;
begin
  if Token = '(' then begin
    Next;
    BoolExpression;
    MatchString(')');
  end
  else begin
    savedtoken := Token;
    savedvalue := Value;
    if (Look = '(') and (DataType(savedvalue) = TOK_PROCEDURE) then
    begin
      idx := Locate(savedvalue);
      rdt := GS_ReturnType[idx];
      if (rdt <> #0) and (rdt <> TOK_STRINGDEF) then
        DoCall(savedvalue)
      else
        AbortMsg('Invalid return type');
    end
    else
    begin
      Next;
      case savedtoken of
        TOK_IDENTIFIER : begin
          if Token = '[' then
          begin
            Next;
            BoolExpression;
            MatchString(']');
            fCCSet := False;
            EmitLn(Format('index %0:s, %1:s, %0:s',[RegisterName, savedvalue]));
          end
          else if savedvalue = 'true' then
            LoadConst('1')
          else if savedvalue = 'false' then
            LoadConst('0')
          else if IsAPIFunc(savedvalue) then
          begin
            DoCallAPIFunc(savedvalue);
          end
          else
            LoadVar(savedvalue);
        end;
        TOK_NUM, TOK_HEX : begin
          LoadConst(savedvalue);
        end;
      else
        Expected('Math Factor');
      end;
    end;
  end;
end;

{--------------------------------------------------------------}
{ Recognize and Translate a Multiply }

procedure TNXCComp.Multiply;
begin
  Next;
  Factor;
  PopMul;
end;

{-------------------------------------------------------------}
{ Recognize and Translate a Divide }

procedure TNXCComp.Divide;
begin
  Next;
  Factor;
  PopDiv;
end;

{-------------------------------------------------------------}
{ Recognize and Translate a Module }

procedure TNXCComp.Modulo;
begin
  Next;
  Factor;
  PopMod;
end;

{---------------------------------------------------------------}
{ Parse and Translate a Math Term }

procedure TNXCComp.Term;
begin
  Factor;
  while IsMulop(Token) do begin
    PushPrim;
    case Token of
      '*': Multiply;
      '/': Divide;
      '%': Modulo;
    end;
  end;
end;

{--------------------------------------------------------------}
{ Recognize and Translate an Add }

procedure TNXCComp.Add;
begin
  Next;
  Term;
  PopAdd;
end;

{-------------------------------------------------------------}
{ Recognize and Translate a Subtract }

procedure TNXCComp.Subtract;
begin
  Next;
  Term;
  PopSub;
end;

{---------------------------------------------------------------}
{ Parse and Translate an Expression }

procedure TNXCComp.Expression;
begin
  if IsAddOp(Token) then
    ClearReg
  else
    Term;
  while IsAddop(Token) do begin
    PushPrim;
    case Token of
      '+': Add;
      '-': Subtract;
    end;
  end;
end;

{---------------------------------------------------------------}
{ Parse and Translate a String Expression }

procedure TNXCComp.StringConcatAssignment(const Name: string);
begin
  if Look = '=' then
  begin
    Next; // move to '='
    Next; // move to next token
    StringExpression(Name, True);
    StoreString(Name);
  end
  else
    AbortMsg('invalid string assignment');
end;

function TNXCComp.ApplyDecoration(const pre, val: string): string;
begin
  Result := Format('__%s_%s', [pre, val]);
end;

function TNXCComp.GetDecoratedIdent(const val : string) : string;
begin
  Result := val;
  if not AlreadyDecorated(val) then
  begin
    case WhatIs(val) of
      stParam, stLocal : begin
        Result := ApplyDecoration(fCurrentThreadName, val);
      end;
    else
      Result := val;
    end;
  end;
end;

function TNXCComp.GetDecoratedValue : string;
begin
  Result := GetDecoratedIdent(Value);
end;

procedure TNXCComp.StringExpression(const Name : string; bAdd : boolean);
var
  asmStr, val : string;
  dt, rdt : char;
  idx : integer;
begin
  fCCSet := False;
  if Look = '(' then
  begin
    // a function call that returns a string
    val := Value;
    if DataType(val) = TOK_PROCEDURE then
    begin
      idx := Locate(val);
      rdt := GS_ReturnType[idx];
      if rdt = TOK_STRINGDEF then
        DoCall(val)
      else
        Expected(sStringReturnValue);
    end
    else
    begin
      Next; // move to '('
      StringFunction(val);
    end;
    if bAdd then
      asmStr := Format('strcat %s, %s, %s', [StrBufName, GetDecoratedIdent(Name), StrRetValName])
    else
      asmStr := Format('mov %s, %s', [StrBufName, StrRetValName]);
    EmitLn(asmStr);
  end
  else if Value = 'asm' then
  begin
    // asm
    Next;
    dt := #0;
    DoAsm(dt);
    if dt <> TOK_STRINGDEF then
      Expected(sStringReturnValue)
    else
    begin
      if bAdd then
        asmStr := Format('strcat %s, %s, %s', [StrBufName, GetDecoratedIdent(Name), StrRetValName])
      else
        asmStr := Format('mov %s, %s', [StrBufName, StrRetValName]);
      EmitLn(asmStr);
    end;
  end
  else
  begin
    CheckString;
    if bAdd then
      asmStr := Format('strcat %s, %s, %s', [StrBufName, GetDecoratedIdent(Name), GetDecoratedValue])
    else
      asmStr := Format('strcat %s, %s', [StrBufName, GetDecoratedValue]);
    Next;
    while Token = '+' do begin
      Next;
      CheckString;
      asmStr := asmStr + ', ' + GetDecoratedValue;
      Next;
    end;
    EmitLn(asmStr);
  end;
end;

{---------------------------------------------------------------}
{ Recognize and Translate a Relational "Equals" }

procedure TNXCComp.Equal;
begin
  Next; // two equal signs of equality comparison
  MatchString('=');
  Expression;
  PopCmpEqual;
  StoreZeroFlag;
end;

procedure TNXCComp.EqualString;
begin
  Next; // two equal signs of equality comparison
  MatchString('=');
  StringExpression('');
  CmpStringHelper('EQ');
  StoreZeroFlag;
end;

procedure TNXCComp.LessString;
begin
  Next;
  case Token of
    '=' : begin
      Next;
      StringExpression('');
      CmpStringHelper('LTEQ');
    end;
    '>' : begin
      Next;
      StringExpression('');
      CmpStringHelper('NEQ');
    end;
  else
    StringExpression('');
    CmpStringHelper('LT');
  end;
  StoreZeroFlag;
end;

procedure TNXCComp.GreaterString;
begin
  Next;
  case Token of
    '=' : begin
      Next;
      StringExpression('');
      CmpStringHelper('GTEQ');
    end;
  else
    StringExpression('');
    CmpStringHelper('GT');
  end;
  StoreZeroFlag;
end;

procedure TNXCComp.NEqualString;
begin
  Next;
  if Token = '=' then
  begin
    Next;
    StringExpression('');
    CmpStringHelper('NEQ');
    StoreZeroFlag;
  end
  else
    Expected('"!="');
end;

{---------------------------------------------------------------}
{ Recognize and Translate a Relational "Less Than or Equal" }

procedure TNXCComp.LessOrEqual;
begin
  Next;
  Expression;
  PopCmpLessOrEqual;
  StoreZeroFlag;
end;

{---------------------------------------------------------------}
{ Recognize and Translate a Relational "Not Equals" }

procedure TNXCComp.NotEqual;
begin
  Next;
  Expression;
  PopCmpNEqual;
  StoreZeroFlag;
end;

{---------------------------------------------------------------}
{ Recognize and Translate a left shift }

procedure TNXCComp.LeftShift;
begin
  Next;
  Expression;
  PopLeftShift;
end;

{---------------------------------------------------------------}
{ Recognize and Translate a right shift }

procedure TNXCComp.RightShift;
begin
  Next;
  Expression;
  PopRightShift;
end;

{---------------------------------------------------------------}
{ Recognize and Translate a Relational "Less Than" }

procedure TNXCComp.Less;
begin
  Next;
  case Token of
    '=' : LessOrEqual;
    '>' : NotEqual;
    '<' : LeftShift;
  else
    Expression;
    PopCmpLess;
    StoreZeroFlag;
  end;
end;

{---------------------------------------------------------------}
{ Recognize and Translate a Relational "Greater Than" }

procedure TNXCComp.Greater;
begin
  Next;
  case Token of
    '=' : begin
      Next;
      Expression;
      PopCmpGreaterOrEqual;
      StoreZeroFlag;
    end;
    '>' : begin
      RightShift;
    end;
  else
    Expression;
    PopCmpGreater;
    StoreZeroFlag;
  end;
end;

procedure TNXCComp.NEqual;
begin
  Next;
  if Token = '=' then
    NotEqual
  else
    Expected('"!="');
end;

{---------------------------------------------------------------}
{ Parse and Translate a Relation }

procedure TNXCComp.Relation;
begin
  if ValueIsStringType then
  begin
    StringExpression('');
    if IsRelop(Token) then begin
      // copy to temp string buffer
      EmitLn(Format('mov %s, %s', [StrTmpBufName, StrBufName]));
      case Token of
        '=': EqualString;
        '<': LessString;
        '>': GreaterString;
        '!': NEqualString;
      end;
    end
    else
      LoadConst('1'); // a string expression is "true"
  end
  else
  begin
    Expression;
    if IsRelop(Token) then begin
      PushPrim;
      case Token of
        '=': Equal;
        '<': Less;
        '>': Greater;
        '!': NEqual;
      end;
    end;
  end;
end;

procedure TNXCComp.StoreZeroFlag;
begin
  EmitLn(Format('mov %s, %s', [RegisterName, ZeroFlag]));
end;

{---------------------------------------------------------------}
{ Parse and Translate a Boolean Factor with Leading NOT }

procedure TNXCComp.NotFactor;
begin
  if Token = '!' then begin
    Next;
    Relation;
    NotIt;
  end
  else
    Relation;
end;

{---------------------------------------------------------------}
{ Parse and Translate a Boolean Term }

procedure TNXCComp.BoolTerm;
begin
  NotFactor;
  while Token = '&' do begin
    PushPrim;
    Next;
    if Token = '&' then
      Next;
    NotFactor;
    PopAnd;
  end;
end;

{--------------------------------------------------------------}
{ Recognize and Translate a Boolean OR }

procedure TNXCComp.BoolOr;
begin
  Next;
  if Token = '|' then
    Next;
  BoolTerm;
  PopOr;
end;

{--------------------------------------------------------------}
{ Recognize and Translate an Exclusive Or }

procedure TNXCComp.BoolXor;
begin
  Next;
  BoolTerm;
  PopXor;
end;

{---------------------------------------------------------------}
{ Parse and Translate a Boolean Expression }

procedure TNXCComp.BoolExpression;
var
  L1, L2 : string;
begin
  fCCSet := False;
  BoolSubExpression;
  while Token = '?' do begin
    // we are parsing a ?: expression
    Next;
    L1 := NewLabel;
    L2 := NewLabel;
    BranchFalse(L1);
    BoolExpression;
    Branch(L2);
    MatchString(':');
    PostLabel(L1);
    BoolExpression;
    PostLabel(L2);
  end;
end;

procedure TNXCComp.BoolSubExpression;
begin
  BoolTerm;
  while IsOrOp(Token) do begin
    PushPrim;
    case Token of
      '|': BoolOr;
      '^': BoolXor;
    end;
  end;
end;

(*
procedure TNXCComp.StringConst;
var
  LitName : string;
  bIsNull : boolean;
begin
  bIsNull := False;
  if NumConsts >= MAXCONSTS then
    AbortMsg('Constants Table is full');
  LitName := Format('CS%d', [CCount]);
  if IsKnown(LitName) <> 0 then
    Duplicate(LitName);
  Value := '';
  if Look = '"' then
    bIsNull := True
  else
  begin
    repeat
      Value := Value + Look;
      GetChar;
    until (Look = '"') or (Look = #0) or endofallsource;
    CS_Value[NumConsts] := Value;
  end;
  CS_Name[NumConsts] := LitName;
  if bIsNull then begin
    CS_Type[NumConsts] := TOK_BYTEDEF;
    CS_Value[NumConsts] := '0';
  end
  else begin
    CS_Type[NumConsts] := TOK_CHARDEF;
  end;
  Next;
  inc(NumConsts);
  inc(CCount);
  // output assembly ?
  EmitLn('; mov __StrBuf,'+LitName);
  Next;
end;
*)

function TNXCComp.GetParamName(procname: string; idx: integer): string;
var
  key : string;
  i : integer;
begin
  Result := '';
  key := Format('%s_%2.2d', [procname, idx]);
  i := fParamNames.IndexOfName(key);
  if i <> -1 then
    Result := ApplyDecoration(procname, fParamNames.ValueFromIndex[i]);
end;

procedure TNXCComp.AddParamName(procname, varname: string; idx: integer);
var
  key : string;
begin
  key := Format('%s_%2.2d', [procname, idx]);
  fParamNames.Add(key+'='+varname);
end;

procedure TNXCComp.DoCall(procname : string);
var
  procpos, protocount, acount : integer;
  dt, rdt : char;
  parname : string;
  bError : boolean;
begin
  acount := 0;
  procpos := Locate(procname);
  protocount := GS_ParamCount[procpos];
  Next;
  bError := Value <> '(';
  if not bError then
    Next
  else
    Expected('"("');
  rdt := GS_ReturnType[procpos];
  while not bError and (Token <> ')') do begin
    if acount >= protocount then
    begin
      AbortMsg('Too many parameters');
      bError := True;
    end;
    dt := GS_ParamType[procpos, acount];
    parname := GetParamName(procname, acount);
    if dt = TOK_STRINGDEF then
    begin
      StringExpression(parname);
      EmitLn(Format('mov %s, %s', [parname, StrBufName]));
    end
    else if dt <> #0 then
    begin
      BoolExpression;
      EmitLn(Format('mov %s, %s', [parname, RegisterName]));
    end;
    inc(acount);
    Scan;
    if Token = ',' then begin
      Next;
      Scan;
    end;
  end;
  if protocount > acount then
    AbortMsg('Too few parameters');
  if Value = ')' then
  begin
    Next;
    EmitLn('call '+procname);
    fCCSet := False;
    if rdt = TOK_STRINGDEF then
    begin
      // copy value from subroutine to register
      EmitLn(Format('mov %s, %s', [StrRetValName, StrBufName(procname)]));
    end
    else if rdt <> #0 then
    begin
      // copy value from subroutine to register
      EmitLn(Format('mov %s, %s', [RegisterName, RegisterName(procname)]));
    end;
  end
  else
    Expected('")"');
  Semi;
  Scan;
end;


procedure TNXCComp.ArrayAssignment(const name : string; dt : char);
var
  tmp, retval : string;
begin
  Next;
  BoolExpression;
  MatchString(']');
  push;
  tmp := tos;
  EmitLn(Format('mov %s, %s', [tmp, RegisterName]));
  MatchString('=');
  if dt = TOK_STRINGDEF then
  begin
    // name of array variable is of type string
    StringExpression(name);
    retval := StrBufName;
  end
  else
  begin
    BoolExpression;
    retval := RegisterName;
  end;
  // move RHS to array[idx]
  EmitLn(Format('replace %0:s, %0:s, %s, %s', [name, tmp, retval]));
  pop;
end;

procedure TNXCComp.MathAssignment(const name : string);
var
  savedtoken : char;
begin
  // Look has to be '=', '+', or '-' or it's all messed up
  if Look = '=' then
  begin
    savedtoken := Token;
    Next; // move to '='
    Next; // move to next token
    BoolExpression;
    case savedtoken of
      '+' : StoreAdd(name);
      '-' : StoreSub(name);
      '*' : StoreMul(name);
      '/' : StoreDiv(name);
      '%' : StoreMod(name);
      '&' : StoreAnd(name);
      '|' : StoreOr(name);
      '^' : StoreXor(name);
    end;
  end
  else if (Token = '+') and (Look = '+') then
  begin
    Next; // move to second +
    Next;
    Semi;
    StoreInc(name, 1);
  end
  else if (Token = '-') and (Look = '-') then
  begin
    Next; // move to second -
    Next;
    Semi;
    StoreDec(name, 1);
  end
  else if (Token = '+') and (Look = '-') then
  begin
    Next; // move to -
    if Look = '=' then
    begin
      Next; // move to '='
      Next; // move to next token
      BoolExpression;
      StoreSign(name);
    end
    else
      AbortMsg(sInvalidAssignment);
  end
  else if (Token = '|') and (Look = '|') then
  begin
    Next; // move to second |
    if Look = '=' then
    begin
      Next; // move to '='
      Next; // move to next token
      BoolExpression;
      StoreAbs(name);
    end
    else
      AbortMsg(sInvalidAssignment);
  end
  else if ((Token = '>') and (Look = '>')) or ((Token = '<') and (Look = '<')) then
  begin
    savedtoken := Token;
    Next; // move to second > or <
    if Look = '=' then
    begin
      Next; // move to '='
      Next; // move to next token
      BoolExpression;
      StoreShift(savedtoken='>', name);
    end
    else
      AbortMsg(sInvalidAssignment);
  end
  else
    AbortMsg(sInvalidAssignment);
end;

procedure TNXCComp.DoLabel;
var
  lbl : string;
begin
  lbl := Value;
  Next; // the colon
  if Locate(lbl) = 0 then
  begin
    AddEntry(lbl, TOK_LABEL);
    PostLabel(lbl);
  end
  else
    Duplicate(lbl);
  Next;
end;

{--------------------------------------------------------------}
{ Parse and Translate an Assignment Statement }

procedure TNXCComp.Assignment;
var
  Name: string;
  dt : char;
begin
  if (IsParam(Value) = 0) and not InTable(Value) and
     (IsLocal(Value) = 0) and not IsAPIFunc(Value) and
     not IsAPIStrFunc(Value) then
    Undefined(Value);
  Name := Value;
  dt := DataType(Name);
  if dt = TOK_PROCEDURE then begin
    DoCall(Name);
  end
  else if dt = TOK_TASK then begin
    AbortMsg('Invalid use of task name');
    SkipLine;
    Next;
  end
  else if dt = TOK_APIFUNC then begin
    Next;
    DoCallAPIFunc(Name); // functions should set register
  end
  else if dt = TOK_APISTRFUNC then begin
    Next;
    StringFunction(Name); // functions should set register
  end
  else begin
    Next;
    if Token = '[' then
    begin
      ArrayAssignment(Name, dt);
    end
    else if Token in ['+','-','/','*','%','&','|','^','>','<'] then
    begin
      if (Token = '+') and (dt = TOK_STRINGDEF) then
        StringConcatAssignment(Name)
      else
        MathAssignment(Name);
    end
    else
    begin
      MatchString('=');
      if dt = TOK_STRINGDEF then
      begin
        StringExpression(Name);
        StoreString(Name);
      end
      else
      begin
        BoolExpression;
        Store(Name);
      end;
    end;
  end;
end;

{---------------------------------------------------------------}
{ Recognize and Translate an IF Construct }

procedure TNXCComp.DoIf(const lend, lstart : string);
var
  L1, L2: string;
begin
  Next;
  MatchString('(');
  BoolExpression;
  MatchString(')');
  L1 := NewLabel;
  L2 := L1;
  BranchFalse(L1);
  Block(lend, lstart);
  if Token = TOK_ELSE then begin
    Next;
    L2 := NewLabel;
    Branch(L2);
    PostLabel(L1);
    Block(lend, lstart);
  end;
  PostLabel(L2);
end;


{--------------------------------------------------------------}
{ Parse and Translate a WHILE Statement }

procedure TNXCComp.DoWhile;
var
  L1, L2: string;
begin
  Next;
  MatchString('(');
  L1 := NewLabel;
  L2 := NewLabel;
  PostLabel(L1);
  BoolExpression;
  MatchString(')');
  BranchFalse(L2);
  Block(L2, L1);
  Branch(L1);
  PostLabel(L2);
end;

procedure TNXCComp.DoDoWhile;
var
  L1, L2: string;
begin
  Next;
  L1 := NewLabel;
  L2 := NewLabel;
  PostLabel(L1);
  Block(L2, L1);
  MatchString('while');
  MatchString('(');
  BoolExpression;
  MatchString(')');
  BranchFalse(L2);
  Branch(L1);
  PostLabel(L2);
end;

procedure TNXCComp.DoRepeat;
var
  L1, L2: string;
  svar : string;
begin
  Next;
  MatchString('(');
  L1 := NewLabel;
  L2 := NewLabel;
  BoolExpression;
  MatchString(')');
  push;
  svar := tos;
  EmitLn(Format('mov %s, %s',[svar, RegisterName]));
  PostLabel(L1);
  StoreDec(svar);
  EmitLn('brtst LT,' + L2 + ', ' + svar);
  Block(L2, L1);
  Branch(L1);
  PostLabel(L2);
  pop;
end;

procedure TNXCComp.DoSwitch;
var
  L2 : string;
  idx : integer;
begin
  Next;
  MatchString('(');
  fSwitchIsString := ValueIsStringType;
  if fSwitchIsString then
    StringExpression('')
  else
    BoolExpression;
  MatchString(')');
  L2 := NewLabel;
  idx := SwitchFixupIndex;
  fInSwitch := True;
  try
    SwitchFixups.Clear;
    Block(L2);
  finally
    fInSwitch := False;
  end;
  PostLabel(L2);
  FixupSwitch(idx, L2);
end;

function TNXCComp.GetCaseConstant: string;
begin
  Result := '';
  // collect tokens up to ':' (this allows for constant expressions)
  while (Token <> ':') and not endofallsource do
  begin
    Result := Result + Value;
    Next;
  end;
  // convert true|false to TRUE|FALSE
  if (Result = 'true') or (Result = 'false') then
    Result := UpperCase(Result);
  // convert 'x' (character constant) into ordinal value
  if (Length(Result) = 3) and
     IsDelimiter('''', Result, 1) and
     IsDelimiter('''', Result, 3) then
  begin
    Result := Copy(Result, 2, 1);
    Result := IntToStr(Ord(Result[1]));
  end;
end;

procedure TNXCComp.DoSwitchCase;
var
  L1 : string;
  caseval, stackval : string;
begin
  caseval := '';
  if fInSwitch then
  begin
    Next; // move past 'case'
    caseval := GetCaseConstant;
    MatchString(':'); // token should be ':' at this point
    L1 := NewLabel;
    PostLabel(L1);
    if fSwitchIsString then
      stackval := StrBufName
    else
      stackval := RegisterName;
    SwitchFixups.Add(Format('brcmp EQ, %s, %s, %s', [L1, caseval, stackval]));
  end
  else
    AbortMsg('case is invalid outside a switch statement');
end;

procedure TNXCComp.DoSwitchDefault;
var
  L1 : string;
begin
  if fInSwitch then
  begin
    Next; // move past 'default'
    MatchString(':');
    L1 := NewLabel;
    PostLabel(L1);
    SwitchFixups.Add(Format('jmp %s', [L1]));
  end
  else
    AbortMsg('default is invalid outside a switch statement');
end;

procedure TNXCComp.FixupSwitch(idx : integer; lbl : string);
var
  i : integer;
begin
  if SwitchFixups.Count = 0 then
    SwitchFixups.Add(Format('jmp %s', [lbl]));
  for i := 0 to SwitchFixups.Count - 1 do
    fNBCSrc.Insert(idx+i, SwitchFixups[i]);
end;

function TNXCComp.SwitchFixupIndex: integer;
begin
  Result := fNBCSrc.Count;
end;

function TNXCComp.ReplaceTokens(const line: string) : string;
begin
  Result := line; // line is already trimmed
  if Length(Result) = 0 then Exit;
  Result := Replace(Result, '__RETURN__', Format(#13#10'mov %s,', [RegisterName]));
  Result := Replace(Result, '__TMPBYTE__', TempByteName);
  Result := Replace(Result, '__TMPWORD__', TempWordName);
  Result := Replace(Result, '__TMPLONG__', TempLongName);
  Result := Replace(Result, '__RETVAL__', RegisterName);
  Result := Replace(Result, '__STRRETVAL__', StrRetValName);
end;

function TNXCComp.DecorateVariables(const asmStr: string): string;
var
  Lex : TGenLexer;

  procedure AddToResult;
  begin
    if Lex.Id = piIdent then
    begin
      // is this a local variable or a parameter?
      Result := Result + GetDecoratedIdent(Lex.Token);
    end
    else
      Result := Result + Lex.Token;
  end;
begin
  Result := '';
  Lex := TNBCLexer.CreateLexer;
  try
    Lex.SetStartData(@asmStr[1], Length(asmStr));
    while not Lex.AtEnd do
    begin
      AddToResult;
      Lex.Next;
    end;
    if Lex.Id <> piUnknown then
      AddToResult;
  finally
    Lex.Free;
  end;
end;

procedure TNXCComp.DoAsm(var dt : char);
var
  asmStr : string;
begin
// gather everything within asm block and output it
  EmitPoundLine;
  MatchString('{');
  asmStr := Value + Look;
  repeat
    repeat
      GetCharX;
      if Look <> '}' then
        asmStr := asmStr + Look;
    until (Look = LF) or (Look = '}') or endofallsource;
    if Pos('__STRRETVAL__', asmStr) > 0 then
      dt := TOK_STRINGDEF
    else
      dt := TOK_LONGDEF;
    asmStr := ReplaceTokens(Trim(asmStr));
    asmStr := DecorateVariables(asmStr);
    if (asmStr <> '') or (Look <> '}') then
      EmitAsmLines(asmStr);
//      EmitLnNoTab(asmStr);
    asmStr := '';
  until (Look = '}') or endofallsource;
  GetChar; // get the '}'
  Next;
end;

{--------------------------------------------------------------}
{ Parse and Translate a FOR Statement }

procedure TNXCComp.DoFor;
var
  L1, L2, L3, L4: string;
begin
  Next;
  MatchString('(');
  L1 := NewLabel;
  L2 := NewLabel;
  L3 := NewLabel;
  L4 := NewLabel;
  Assignment;
  while Token = ',' do
  begin
    Next;
    Assignment;
  end;
  Semi;
  PostLabel(L1);
  BoolExpression;
  Semi;
  BranchFalse(L2);
  Branch(L3);
  PostLabel(L4);
  Assignment;
  while Token = ',' do
  begin
    Next;
    Assignment;
  end;
  Semi;
  MatchString(')');
  Branch(L1);
  PostLabel(L3);
  Block(L2, L4);
  Branch(L4);
  PostLabel(L2);
end;

{--------------------------------------------------------------}
{ Process a wait Statement }

procedure TNXCComp.DoWait;
begin
  Next;
  MatchString('(');
  BoolExpression;
  MatchString(')');
  EmitLn(Format('waitv %s',[RegisterName]));
end;

function IndexOfAPICommand(const name : string) : integer;
begin
  for Result := Low(APIList) to High(APIList) do
  begin
    if APIList[Result] = name then
      Exit;
  end;
  Result := -1;
end;

procedure TNXCComp.DoAPICommands(const lend, lstart : string);
var
  idx : integer;
begin
  idx := IndexOfAPICommand(Value);
  case idx of
    API_BREAK    : DoBreakContinue(idx, lend);
    API_CONTINUE : DoBreakContinue(idx, lstart);
    API_RETURN   : DoReturn;
    API_WAIT     : DoWait;
    API_ONFWD,
    API_ONREV    : DoOnFwdRev;
    API_ONFWDREG,
    API_ONREVREG : DoOnFwdRevReg;
    API_OFF,
    API_COAST,
    API_FLOAT    : DoStopMotors;
    API_ONFWDSYNC,
    API_ONREVSYNC : DoOnFwdRevSync;
    API_ROTATEMOTOR,
    API_ROTATEMOTOREX : DoRotateMotors(idx);
    API_SETSENSORTYPE,
    API_SETSENSORMODE : DoSetSensorTypeMode(idx);
    API_CLEARSENSOR,
    API_SETSENSORTOUCH,
    API_SETSENSORLIGHT,
    API_SETSENSORSOUND,
    API_SETSENSORLOWSPEED,
    API_RESETSENSOR : DoClearSetResetSensor;
    API_PRECEDES,
    API_FOLLOWS : DoPrecedesFollows;
    API_ACQUIRE,
    API_RELEASE : DoAcquireRelease;
    API_EXITTO : DoExitTo;
    API_SETINPUT,
    API_SETOUTPUT : DoSetInputOutput(idx);
    API_STOP : DoStop;
    API_GOTO : DoGoto;
  else
    AbortMsg('Unknown API command');
  end;
end;

{--------------------------------------------------------------}
{ Parse and Translate a Single Statement }

procedure TNXCComp.Statement(const lend, lstart : string);
var
  dt : Char;
begin
  if Token = '{' then
    Block(lend, lstart)
  else
  begin
    ProcessDirectives;
    case Token of
      TOK_IF:         DoIf(lend, lstart);
      TOK_WHILE:      DoWhile;
      TOK_FOR:        DoFor;
      TOK_DO:         DoDoWhile;
      TOK_REPEAT:     DoRepeat;
      TOK_SWITCH:     DoSwitch;
      TOK_CASE:       DoSwitchCase;
      TOK_DEFAULT:    DoSwitchDefault;
      TOK_ASM: begin
        Next;
        dt := #0;
        DoAsm(dt);
      end;
      TOK_API:        DoAPICommands(lend, lstart);
      TOK_IDENTIFIER: begin
        if Look = ':' then
          DoLabel
        else
          Assignment;
      end;
      ')', ';' : Next;
    end;
    EmitPoundLine;
  end;
end;


{--------------------------------------------------------------}
{ Parse and Translate a Block of Statements }

procedure TNXCComp.Block(const lend, lstart : string);
var
  bBlockStatement : boolean;
begin
  bBlockStatement := Value = '{';
  if bBlockStatement then
  begin
    Next;
    BlockStatements(lend, lstart);
    MatchString('}');
    Scan;
  end
  else
  begin
    Scan;
    Statement(lend, lstart);
    Semi;
    Scan;
  end;
end;

procedure TNXCComp.CheckBytesRead(const oldBytesRead: integer);
begin
  if fBytesRead = oldBytesRead then
  begin
    AbortMsg('parser error');
    SkipLine;
    Next;
  end;
end;

procedure TNXCComp.BlockStatements(const lend, lstart: string);
var
  oldBytesRead : integer;
begin
  Scan;
  while not (Token in [TOK_END, TOK_ELSE]) and not endofallsource do
  begin
    oldBytesRead := fBytesRead;
    Statement(lend, lstart);
    Semi;
    Scan;
    CheckBytesRead(oldBytesRead);
  end;
end;

{--------------------------------------------------------------}
{ Allocate Storage for a Variable }

procedure TNXCComp.AllocLocal(const sub : string; dt : char);
var
  savedval : string;
  ival, aval : string;
begin
  Next;
  if Token <> TOK_IDENTIFIER then Expected('Variable Name');
  AddLocal(ApplyDecoration(sub, Value), dt);
  savedval := Value;
  ival := '';
  Next;
  aval := '';
  if (Token = '[') and (Look = ']') then begin
    // declaring an array
    while Token in ['[', ']'] do begin
      aval := aval + Token;
      Next;
    end;
    dt := ArrayOfType(dt);
  end;
  if (Token = ',') or (Token = ';') then
  begin
    Allocate(ApplyDecoration(sub, savedval), aval, ival, dt);
  end
  else if Token = '=' then
  begin
    ival := GetInitialValue(dt);
    Allocate(ApplyDecoration(sub, savedval), aval, ival, dt);
  end
  else
    Next;
end;

function TNXCComp.GetInitialValue(dt : char): string;
begin
  Result := '';
  // handle string variables differently
  Next; // move to first token past the '='
  if dt = TOK_STRINGDEF then
  begin
    Result := Value;
    if Token <> TOK_STRINGLIT then
      AbortMsg('invalid string initialization');
    Next;
  end
  else
  begin
    while not (Token in [',', ';']) and not endofallsource do
    begin
      Result := Result + Value;
      Next;
    end;
  end;
end;

procedure TNXCComp.AllocGlobal(dt : char);
var
  savedval : string;
  ival : string;
  aval : string;
begin
  Next;
  if Token <> TOK_IDENTIFIER then Expected('Variable Name');
  CheckDup(Value);
  // optional initial value
  savedval := Value;
  ival := '';
  Next;
  // it is possible that we are looking at a function declaration
  // rather than a variable declaration.
  if Token = '(' then
  begin
    FunctionBlock(savedval, dt);
  end
  else
  begin
    aval := '';
    if (Token = '[') and (Look = ']') then begin
      // declaring an array
      while Token in ['[', ']'] do begin
        aval := aval + Token;
        Next;
      end;
      dt := ArrayOfType(dt);
    end;
    AddEntry(savedval, dt);
    if (Token = ',') or (Token = ';') then
    begin
      Allocate(savedval, aval, ival, dt);
    end
    else if Token = '=' then
    begin
      ival := GetInitialValue(dt);
      Allocate(savedval, aval, ival, dt);
//      Next;
    end
    else
      Next;
  end;
end;

function TNXCComp.GetVariableType(vt : char; bUnsigned : boolean) : char;
begin
  if not bUnsigned then
    Result := vt
  else
    case vt of
      TOK_LONGDEF : Result := TOK_ULONGDEF;
      TOK_SHORTDEF : Result := TOK_USHORTDEF;
      TOK_CHARDEF : Result := TOK_BYTEDEF;
    else
      Result := vt;
    end;
end;

{--------------------------------------------------------------}
{ Parse and Translate Global Declarations }

procedure TNXCComp.TopDecls;
var
  vt : char;
  bUnsigned : boolean;
  oldBytesRead : Integer;
begin
  bUnsigned := False;
  Scan;
  while not (Token in [TOK_TASK, TOK_PROCEDURE]) and not endofallsource do
  begin
    oldBytesRead := fBytesRead;
    case Token of
      TOK_DIRECTIVE : begin
        ProcessDirectives;
        Scan;
      end;
      TOK_UNSIGNED : begin
        Next;
        Scan;
        bUnsigned := True;
      end;
      TOK_LONGDEF, TOK_SHORTDEF,
      TOK_CHARDEF, TOK_BYTEDEF,
      TOK_MUTEXDEF, TOK_STRINGDEF : begin
        vt := Token;
        AllocGlobal(GetVariableType(vt, bUnsigned));
        while Token = ',' do
          AllocGlobal(GetVariableType(vt, bUnsigned));
        Semi;
        Scan;
        bUnsigned := False;
      end;
    else
      // nothing here right now
      Semi;
      Scan;
    end;
    CheckBytesRead(oldBytesRead);
  end;
end;

procedure TNXCComp.AddLocal(name : string; dt : char);
begin
  if (IsParam(name) <> 0) or (IsLocal(name) <> 0) then
    Duplicate(name);
  Inc(NumLocals);
  LS_Name[NumLocals] := name;
  LS_Type[NumLocals] := dt;
end;

procedure TNXCComp.DoLocals(const sub : string);
var
  bIsUnsigned : boolean;
  dt : char;
begin
  bIsUnsigned := False;
  ClearLocals;
  Scan;
  while (Token in [TOK_DIRECTIVE, TOK_UNSIGNED, TOK_LONGDEF, TOK_SHORTDEF,
    TOK_CHARDEF, TOK_BYTEDEF, TOK_MUTEXDEF, TOK_STRINGDEF]) and not endofallsource do
  begin
    case Token of
      TOK_DIRECTIVE : begin
        ProcessDirectives;
        Scan;
      end;
      TOK_UNSIGNED : begin
        Next;
        Scan;
        bIsUnsigned := True;
      end;
      TOK_LONGDEF, TOK_SHORTDEF,
      TOK_CHARDEF, TOK_BYTEDEF,
      TOK_MUTEXDEF, TOK_STRINGDEF : begin
        dt := Token;
        AllocLocal(sub, GetVariableType(dt, bIsUnsigned));
        while Token = ',' do
          AllocLocal(sub, GetVariableType(dt, bIsUnsigned));
        Semi;
        Scan;
        bIsUnsigned := False;
      end;
    else
      Expected('Valid Program Block Statement');
    end;
  end;
end;

function TNXCComp.FormalList(protoexists : boolean; procname : string) : integer;
var
  protocount : integer;
  pltype : integer;
  pcount, procpos : integer;
  ptype : char;
  varnam : string;
  bIsUnsigned : boolean;
  aval : string;
const
  HASPROTO = 2;
  HASNOPROTO = 3;
begin
  protocount := 0;
  pcount := 0;
  pltype := 0;
  ptype := #0;
  procpos := Locate(procname);
  if protoexists then
    protocount := GS_ParamCount[procpos];
  while Token <> ')' do begin
    Scan;
    if Token = TOK_UNSIGNED then begin
      bIsUnsigned := True;
      Next;
      Scan;
      pltype := 1;
    end
    else
      bIsUnsigned := False;
    if Token in [TOK_CHARDEF, TOK_BYTEDEF, TOK_SHORTDEF, TOK_LONGDEF, TOK_MUTEXDEF, TOK_STRINGDEF] then
    begin
      if protoexists then
        Expected('Parameter List');
      ptype := Token;
      pltype := 1;
      Next;
      Scan;
      if (Token <> '[') and (Token <> ',') and (Token <> ')') and (Token <> TOK_IDENTIFIER) then
        AbortMsg('Unexpected character encountered');
    end
    else
      if bIsUnsigned then
        AbortMsg('Missing Data Type');
    if Token = TOK_IDENTIFIER then
    begin
      varnam := Value;
      Next;
      Scan;
      if pltype = 1 then
        pltype := HASNOPROTO
      else
        pltype := HASPROTO;
    end;
    ptype := GetVariableType(ptype, bIsUnsigned);
    aval := '';
    if (Token = '[') and (Look = ']') then begin
      // declaring an array
      while Token in ['[', ']'] do begin
        aval := aval + Token;
        Next;
      end;
      ptype := ArrayOfType(ptype);
    end;
    case pltype of
      1 : begin
        GS_ParamType[procpos, protocount] := ptype;
        if protocount >= MAXPARAMS then
          AbortMsg(sMaxParamCountExceeded);
        inc(protocount);
      end;
      HASPROTO : begin
        if not protoexists then
          Expected('Data type');
        if pcount >= MAXPARAMS then
          AbortMsg(sMaxParamCountExceeded);
        AddParam(ApplyDecoration(procname, varnam), GS_ParamType[procpos, pcount]);
        inc(pcount);
        if pcount > protocount then
          AbortMsg(sTooManyArgs);
      end;
      HASNOPROTO : begin
        if protoexists then
          AbortMsg('Data types already defined in prototype');
        GS_ParamType[procpos, protocount] := ptype;
        if protocount >= MAXPARAMS then
          AbortMsg(sMaxParamCountExceeded);
        AddParam(ApplyDecoration(procname, varnam), ptype);
        Allocate(ApplyDecoration(procname, varnam), aval, '', ptype);
        AddParamName(procname, varnam, pcount);
        inc(protocount);
        inc(pcount);
      end;
    end;
    while Token = ',' do begin
      Next;
      Scan;
      if (pltype = 1) or (pltype = 3) then begin
        if Token = TOK_UNSIGNED then begin
          bIsUnsigned := True;
          Next;
          Scan;
        end
        else
          bIsUnsigned := False;
        if Token in [TOK_CHARDEF, TOK_BYTEDEF, TOK_SHORTDEF, TOK_LONGDEF, TOK_MUTEXDEF, TOK_STRINGDEF] then
        begin
          ptype := Token;
          Next;
          Scan;
        end
        else if bIsUnsigned then
          AbortMsg('Missing Data Type');
      end;
      if (pltype = HASPROTO) or (pltype = HASNOPROTO) then begin
        if Token = TOK_IDENTIFIER then begin
          varnam := Value;
          Next;
          Scan;
        end
        else
          Expected(sVariableName);
      end;
      aval := '';
      if (Token = '[') and (Look = ']') then begin
        // declaring an array
        while Token in ['[', ']'] do begin
          aval := aval + Token;
          Next;
        end;
        ptype := ArrayOfType(ptype);
      end;
      case pltype of
        1 : begin
          GS_ParamType[procpos, protocount] := ptype;
          if protocount >= MAXPARAMS then
            AbortMsg(sMaxParamCountExceeded);
          inc(protocount);
        end;
        HASPROTO : begin
          if pcount >= MAXPARAMS then
            AbortMsg(sMaxParamCountExceeded);
          AddParam(ApplyDecoration(procname, varnam), GS_ParamType[procpos, pcount]);
          inc(pcount);
          if pcount > protocount then
            AbortMsg(sTooManyArgs);
        end;
        HASNOPROTO : begin
          GS_ParamType[procpos, protocount] := ptype;
          if protocount >= MAXPARAMS then
            AbortMsg(sMaxParamCountExceeded);
          AddParam(ApplyDecoration(procname, varnam), ptype);
          Allocate(ApplyDecoration(procname, varnam), aval, '', ptype);
          AddParamName(procname, varnam, pcount);
          inc(protocount);
          inc(pcount);
        end;
      end;
    end; // while Token = ','
  end; // while Token <> ')'
  if protoexists and (pcount <> protocount) then
    AbortMsg('Too few arguments');
  case pltype of
    0 : begin
      // nothing
    end;
    1 : begin
      GS_ParamCount[procpos] := protocount;
    end;
    HASPROTO : begin
      GS_ParamCount[procpos] := pcount;
    end;
    HASNOPROTO : begin
      GS_ParamCount[procpos] := pcount;
    end;
  end;
  Result := pltype;
end;

procedure TNXCComp.ProcedureBlock;
var
  Name : string;
  procexists : integer;
  protoexists, bIsSub : boolean;
  pltype : integer;
  savedToken : char;
begin
  while Token in [TOK_PROCEDURE, TOK_TASK] do
  begin
    bIsSub := Token = TOK_PROCEDURE;
    savedToken := Token;
    Next;
    CheckIdent;
    Name := Value;
    procexists := Locate(Name);
    protoexists := False;
    if procexists <> 0 then begin
      if not (GS_Type[procexists] in [TOK_PROCEDURE, TOK_TASK]) then
        Duplicate(Name);
      if GS_Size[procexists] = 0 then
        protoexists := True
      else
        Duplicate(Name);
    end
    else begin
      AddEntry(Name, savedToken);
      GS_ReturnType[NumGlobals] := #0;
    end;
    Next;
    MatchString('(');
    fCurrentThreadName := Name;
    fThreadNames.Add(Name);
    if bIsSub then
      pltype := FormalList(protoexists, Name)
    else
      pltype := 0;
    MatchString(')');
    Semi;
    Scan;
    if Token = TOK_BEGIN then
    begin
      if pltype = 1 then
        AbortMsg('Not valid for a prototype');
      if protoexists then
        GS_Size[procexists] := 1
      else
        GS_Size[NumGlobals] := 1;
      Prolog(Name, bIsSub);
      MatchString('{');
      DoLocals(Name);
      BlockStatements();
      MatchString('}');
      Epilog(bIsSub);
      Scan;
    end
    else
    begin
      if protoexists then
        Expected('Prototype already defined - "begin"');
      Scan;
    end;
    ClearParams;
    TopDecls;
  end;
end;

procedure TNXCComp.FunctionBlock(const Name: string; dt: char);
var
  procexists : integer;
  protoexists : boolean;
  pltype : integer;
begin
  procexists := Locate(Name);
  protoexists := False;
  if procexists <> 0 then begin
    if not (GS_Type[procexists] in [TOK_PROCEDURE, TOK_TASK]) then
      Duplicate(Name);
    if GS_Size[procexists] = 0 then
      protoexists := True
    else
      Duplicate(Name);
  end
  else begin
    AddEntry(Name, TOK_PROCEDURE);
    GS_ReturnType[NumGlobals] := dt;
  end;
  MatchString('(');
  fCurrentThreadName := Name;
  fThreadNames.Add(Name);
  pltype := FormalList(protoexists, Name);
  MatchString(')');
  Semi;
  Scan;
  if Token = TOK_BEGIN then
  begin
    if pltype = 1 then
      AbortMsg('Not valid for a prototype');
    if protoexists then
      GS_Size[procexists] := 1
    else
      GS_Size[NumGlobals] := 1;
    Prolog(Name, True);
    MatchString('{');
    DoLocals(Name);
    BlockStatements();
    MatchString('}');
    Epilog(True);
    Scan;
  end
  else
  begin
    if protoexists then
      Expected('Prototype already defined - "begin"');
    Scan;
  end;
  ClearParams;
end;

{--------------------------------------------------------------}
{ Initialize }

procedure TNXCComp.Init;
var
  i : integer;
begin
  fCurrentLine := '';
  totallines := 1;
  for i := 1 to MAXINCL do
    linenumber[i] := 1;
  ClearParams;
  sd := 0;
  maxsd := 0;
  GetChar;
  Next;
end;

{--------------------------------------------------------------}
{  Parse and Translate a Program }

procedure TNXCComp.Prog;
begin
  Header;
  TopDecls;
  if Token in [TOK_PROCEDURE, TOK_TASK] then
    ProcedureBlock;
  Trailer;
end;

{ TNXCComp }

constructor TNXCComp.Create;
begin
  fTmpAsmLines := TStringList.Create;
  fNBCSrc := TStringList.Create;
  fMS := TMemoryStream.Create;
  fMessages := TStringList.Create;
  fIncludeDirs := TStringList.Create;
  fAPIFunctions := TStringList.Create;
  TStringList(fAPIFunctions).Sorted := True;
  fAPIStrFunctions := TStringList.Create;
  TStringList(fAPIStrFunctions).Sorted := True;
  fThreadNames := TStringList.Create;
  TStringList(fThreadNames).Sorted := True;
  TStringList(fThreadNames).Duplicates := dupIgnore;
  fParamNames := TStringList.Create;
  TStringList(fParamNames).Sorted := True;
  fSwitchFixups := TStringList.Create;
  fInSwitch := False;
  LoadAPIFunctions;
end;

destructor TNXCComp.Destroy;
begin
  FreeAndNil(fTmpAsmLines);
  FreeAndNil(fNBCSrc);
  FreeAndNil(fMS);
  FreeAndNil(fMessages);
  FreeAndNil(fIncludeDirs);
  FreeAndNil(fAPIFunctions);
  FreeAndNil(fAPIStrFunctions);
  FreeAndNil(fThreadNames);
  FreeAndNil(fParamNames);
  FreeAndNil(fSwitchFixups);
  inherited;
end;

procedure TNXCComp.InternalParseStream;
begin
  try
    fParamNames.Clear;
    fThreadNames.Clear;
    fBadProgram     := False;
    fBytesRead      := 0;
    fProgErrorCount := 0;
    PreProcess;
    fMS.Position := 0;
    Init;
    Prog;
  except
    on E : EPreprocessorException do
    begin
      fBadProgram := True;
      ReportProblem(E.LineNo, CurrentFile, E.Message, true);
    end;
    on E : Exception do
    begin
      fBadProgram := True;
      ReportProblem(linenumber[slevel], CurrentFile, E.Message, true);
    end;
  end;
end;

procedure TNXCComp.Parse(aStrings: TStrings);
begin
  Clear;
  aStrings.SaveToStream(fMS);
  InternalParseStream;
end;

procedure TNXCComp.Parse(aStream: TStream);
begin
  Clear;
  fMS.CopyFrom(aStream, 0);
  InternalParseStream;
end;

procedure TNXCComp.Parse(const aFilename: string);
begin
  Clear;
  fMS.LoadFromFile(aFilename);
  InternalParseStream;
end;

procedure TNXCComp.Clear;
begin
  fNBCSrc.Clear;
  fMessages.Clear;
  fTempChar := ' ';
end;

{--------------------------------------------------------------}
{ Recognize and Translate a break/continue }

procedure TNXCComp.DoBreakContinue(idx : integer; const lbl: string);
var
  val : string;
begin
  val := APIList[idx];
  MatchString(val);
  Semi;
  if lbl <> '' then
    Branch(lbl)
  else
    AbortMsg(Format('%s is invalid outside of a loop', [val]));
end;

procedure TNXCComp.DoOnFwdRev;
var
  op, arg1 : string;
begin
  //OnFwd(ports, pwr)
  op := Value;
  Next;
  MatchString('(');
  // ports
  arg1 := GetDecoratedValue;
  Next;
  MatchString(',');
  // pwr
  BoolExpression;
  MatchString(')');
  EmitLn(Format('%s(%s, %s)',[op, arg1, RegisterName]));
end;

procedure TNXCComp.DoOnFwdRevReg;
var
  op, arg1, svar : string;
begin
  //OnFwdReg(ports, pwr, regmode)
  op := Value;
  Next;
  MatchString('(');
  // ports
  arg1 := GetDecoratedValue;
  Next;
  MatchString(',');
  // pwr
  BoolExpression;
  push;
  svar := tos;
  EmitLn(Format('mov %s, %s',[svar, RegisterName]));
  MatchString(',');
  // regmode
  BoolExpression;
  MatchString(')');
  EmitLn(Format('%s(%s, %s, %s)',[op, arg1, svar, RegisterName]));
  pop;
end;

procedure TNXCComp.DoOnFwdRevSync;
var
  op, ports, pwr : string;
begin
  //OnFwdSync(ports, pwr, turnpct)
  op := Value;
  Next;
  MatchString('(');
  // ports
  ports := GetDecoratedValue;
  Next;
  MatchString(',');
  // pwr
  BoolExpression;
  push;
  pwr := tos;
  EmitLn(Format('mov %s, %s',[pwr, RegisterName]));
  MatchString(',');
  // turnpct
  BoolExpression;
  MatchString(')');
  EmitLn(Format('%s(%s, %s, %s)',[op, ports, pwr, RegisterName]));
  pop;
end;

procedure TNXCComp.DoRotateMotors(idx: integer);
var
  op, ports, pwr, angle, turnpct : string;
begin
 //RotateMotor(ports, pwr, angle)
 //RotateMotorEx(ports, pwr, angle, turnpct, bSync)
  op := Value;
  Next;
  MatchString('(');
  // ports
  ports := GetDecoratedValue;
  Next;
  MatchString(',');
  // power
  BoolExpression;
  push;
  pwr := tos;
  EmitLn(Format('mov %s, %s',[pwr, RegisterName]));
  MatchString(',');
  // angle
  BoolExpression;
  if idx = API_ROTATEMOTOR then
  begin
    MatchString(')');
    EmitLn(Format('%s(%s, %s, %s)', [op, ports, pwr, RegisterName]))
  end
  else
  begin
    push;
    angle := tos;
    EmitLn(Format('mov %s, %s',[angle, RegisterName]));
    MatchString(',');
    // turn pct
    BoolExpression;
    push;
    turnpct := tos;
    EmitLn(Format('mov %s, %s',[turnpct, RegisterName]));
    MatchString(',');
    // sync
    BoolExpression;
    MatchString(')');
    EmitLn(Format('%s(%s, %s, %s, %s, %s)', [op, ports, pwr, angle, turnpct, RegisterName]));
    pop;
    pop;
  end;
  pop;
end;

procedure TNXCComp.DoClearSetResetSensor;
var
  op, port : string;
begin
  //ClearSensor(port)
  //SetSensorTouch(port)
  //SetSensorLight(port)
  //SetSensorSound(port)
  //SetSensorLowspeed(port)
  //ResetSensor(port)
  op := Value;
  Next;
  MatchString('(');
  // port
  port := GetDecoratedValue;
  Next;
  MatchString(')');
  EmitLn(op + '(' + port + ')');
end;

procedure TNXCComp.DoSetSensorTypeMode(idx : integer);
var
  port : string;
begin
  //SetSensorType(port,type)
  //SetSensorMode(port,mode)
  Next;
  MatchString('(');
  // port
  port := GetDecoratedValue;
  Next;
  MatchString(',');
  // type/mode
  BoolExpression;
  MatchString(')');
  if idx = API_SETSENSORTYPE then
    EmitLn(Format('setin %s, %s, Type',[RegisterName, port]))
  else
    EmitLn(Format('setin %s, %s, InputMode',[RegisterName, port]));
end;

procedure TNXCComp.DoAcquireRelease;
var
  op, val : string;
begin
 // Acquire(mutex);
 // Release(mutex);
  op := LowerCase(Value);
  Next;
  MatchString('(');
  // mutex
  val := GetDecoratedValue;
  CheckIdent;
  CheckTable(val);
  if DataType(val) <> TOK_MUTEXDEF then Expected('Mutex');
  Next;
  MatchString(')');
  EmitLn(op + ' ' + val);
end;

procedure TNXCComp.DoPrecedesFollows;
var
  op, val : string;
begin
  // Precedes(x, y, z, ...);
  // Follows(x, y, z);
  op := LowerCase(Value);
  Next;
  MatchString('(');
  val := Value;
  CheckIdent;
  CheckTable(val);
  if DataType(val) <> TOK_TASK then
    Expected(sTaskName);
  Next;
  while Value = ',' do
  begin
    val := val + Value; // the comma
    Next;
    CheckIdent;
    CheckTable(Value);
    val := val + ' ' + Value; // the next value
    Next;
  end;
  MatchString(')');
  EmitLn(op + ' ' + val);
end;

procedure TNXCComp.DoExitTo;
var
  op, val : string;
begin
  // ExitTo(task);
  op := LowerCase(Value);
  Next;
  MatchString('(');
  // task
  val := Value;
  CheckIdent;
  CheckTable(val);
  if DataType(val) <> TOK_TASK then
    Expected(sTaskName);
  Next;
  MatchString(')');
  EmitLn(op + ' ' + val);
end;

procedure TNXCComp.DoStop;
begin
  // Stop(stop?);
  Next;
  MatchString('(');
  // stop?
  BoolExpression;
  MatchString(')');
  EmitLn(Format('stop %s',[RegisterName]));
end;

procedure TNXCComp.DoGoto;
begin
  // goto labelName;
  Next;
  // labelName
  CheckIdent;
  EmitLn(Format('jmp %s',[Value]));
  Next;
end;

procedure TNXCComp.DoSetInputOutput(idx: integer);
var
  port, field : string;
begin
  // SetInput(port, field, value)
  // SetOutput(ports, field, value)
  Next;
  MatchString('(');
  // port
  port := GetDecoratedValue;
  Next;
  MatchString(',');
  // field
  field := Value;
  CheckNumeric;
  Next;
  MatchString(',');
  // value
  BoolExpression;
  MatchString(')');
  case idx of
    API_SETINPUT  :
      EmitLn(Format('setin %s, %s, %s', [RegisterName, port, field]));
    API_SETOUTPUT :
      EmitLn(Format('setout %s, %s, %s', [port, field, RegisterName]));
  end;
end;

procedure TNXCComp.DoReturn;
var
  dt : char;
  idx : integer;
begin
  // return
  idx := Locate(fCurrentThreadName);
  if GS_Type[idx] <> TOK_PROCEDURE then
    AbortMsg('return is invalid within a task');
  dt := GS_ReturnType[idx];
  Next;
  // leave return value on "stack"
  if dt = TOK_STRINGDEF then
    StringExpression('')
  else if dt <> #0 then
    BoolExpression;
  Semi;
  EmitLn('return');
end;

procedure TNXCComp.DoStopMotors;
var
  op, arg1 : string;
begin
  // Off(ports)
  // Coast(ports)
  // Float(ports)
  op := Value;
  Next;
  MatchString('(');
  // ports
  arg1 := GetDecoratedValue;
  Next;
  MatchString(')');
  EmitLn(op + '(' + arg1 + ')');
end;

procedure TNXCComp.PreProcess;
var
  P : TLangPreprocessor;
begin
  P := TLangPreprocessor.Create(TNQCLexer, ExtractFilePath(ParamStr(0)));
  try
    P.AddIncludeDirs(IncludeDirs);
    P.Preprocess(CurrentFile, fMS);
  finally
    P.Free;
  end;
end;

procedure TNXCComp.ProcessDirectives;
begin
  while Token = TOK_DIRECTIVE do
  begin
    Next;
    // look for #line statements
    if (Token = TOK_IDENTIFIER) and (LowerCase(Value) = 'line') then
    begin
      SkipDirectiveLine;
      HandlePoundLine;
      Next;
    end
    else
    begin
      SkipDirectiveLine;
      Next;
    end;
    EmitLn(Trim(fDirLine));
    Scan;
  end;
end;

procedure TNXCComp.HandlePoundLine;
var
  i : integer;
  tmpLine, tmpFile : string;
begin
  i := Pos('#line ', fDirLine);
  if i = 1 then
  begin
    // this is a special preprocessor line
    tmpLine := Trim(fDirLine);
    Delete(tmpLine, 1, 6);
    i := Pos(' ', tmpLine);
    linenumber[slevel] := StrToIntDef(Copy(tmpLine, 1, i - 1), linenumber[slevel]);
    IncLineNumber;
    Delete(tmpLine, 1, i);
    tmpFile      := Replace(tmpLine, '"', '');
    CurrentFile  := tmpFile;
  end;
end;

procedure TNXCComp.IncLineNumber;
begin
  linenumber[slevel] := linenumber[slevel] + 1;
  inc(totallines);
end;

function TNXCComp.APIFuncNameToID(procname: string): integer;
begin
  Result := StrToIntDef(fAPIFunctions.Values[procname], 0);
end;

function TNXCComp.APIStrFuncNameToID(procname: string): integer;
begin
  Result := StrToIntDef(fAPIStrFunctions.Values[procname], 0);
end;

function TNXCComp.IsAPIFunc(procname: string): boolean;
begin
  Result := fAPIFunctions.IndexOfName(procname) <> -1;
end;

function TNXCComp.IsAPIStrFunc(procname: string): boolean;
begin
  Result := fAPIStrFunctions.IndexOfName(procname) <> -1;
end;

procedure TNXCComp.AddAPIFunction(const name: string; id: integer);
begin
  fAPIFunctions.Add(name + '=' + IntToStr(id));
end;

procedure TNXCComp.AddAPIStringFunction(const name: string; id: integer);
begin
  fAPIStrFunctions.Add(name + '=' + IntToStr(id));
end;

function TNXCComp.tos : string;
begin
  Result := Format('__stack_%3.3d%s', [sd, fCurrentThreadName]);
end;

function TNXCComp.TempByteName: string;
begin
  Result := Format('__tmpsbyte%s', [fCurrentThreadName]);
end;

function TNXCComp.TempWordName: string;
begin
  Result := Format('__tmpsword%s', [fCurrentThreadName]);
end;

function TNXCComp.TempLongName: string;
begin
  Result := Format('__tmpslong%s', [fCurrentThreadName]);
end;

function TNXCComp.RegisterName(name : string): string;
begin
  if name = '' then
    name := fCurrentThreadName;
  Result := Format('__D0%s',[name]);
end;

function TNXCComp.ZeroFlag: string;
begin
  Result := Format('__zf%s', [fCurrentThreadName]);
end;

function TNXCComp.StrTmpBufName(name : string): string;
begin
  if name = '' then
    name := fCurrentThreadName;
  Result := Format('__strtmpbuf%s', [name]);
end;

function TNXCComp.StrBufName(name : string): string;
begin
  if name = '' then
    name := fCurrentThreadName;
  Result := Format('__strbuf%s', [name]);
end;

function TNXCComp.StrRetValName(name : string): string;
begin
  if name = '' then
    name := fCurrentThreadName;
  Result := Format('__strretval%s', [name]);
end;

procedure TNXCComp.EmitRegisters;
var
  j : integer;
begin
  for j := 0 to fThreadNames.Count - 1 do
  begin
    EmitLn(Format('__tmpsbyte%s sbyte',[fThreadNames[j]]));
    EmitLn(Format('__tmpsword%s sword',[fThreadNames[j]]));
    EmitLn(Format('__tmpslong%s slong',[fThreadNames[j]]));
    EmitLn(Format('__D0%s slong',[fThreadNames[j]]));
    EmitLn(Format('__zf%s byte',[fThreadNames[j]]));
    EmitLn(Format('__strtmpbuf%s byte[]',[fThreadNames[j]]));
    EmitLn(Format('__strbuf%s byte[]',[fThreadNames[j]]));
    EmitLn(Format('__strretval%s byte[]',[fThreadNames[j]]));
  end;
end;

procedure TNXCComp.EmitStackVariables;
var
  i, j : integer;
begin
  for j := 0 to fThreadNames.Count - 1 do
    for i := 1 to maxsd do
      EmitLn(Format('__stack_%3.3d%s slong', [i, fThreadNames[j]]));
end;

const
  APISF_NUMTOSTR   = 0;
  APISF_STRCAT     = 1;
  APISF_SUBSTR     = 2;
  APISF_FLATTEN    = 3;
  APISF_STRREPLACE = 4;

procedure TNXCComp.StringFunction(const Name : string);
var
  id : integer;
  op : string;
begin
  id := APIStrFuncNameToID(Name);
  case id of
    APISF_NUMTOSTR, APISF_FLATTEN : begin
      MatchString('(');
      BoolExpression;
      MatchString(')');
      if id = APISF_NUMTOSTR then
        op := 'numtostr'
      else
        op := 'flatten';
      EmitLn(Format('%s %s, %s', [op, StrRetValName, RegisterName]));
    end;
    APISF_STRCAT : DoStrCat;
    APISF_SUBSTR : DoSubString;
    APISF_STRREPLACE : DoStrReplace;
  end;
end;

procedure TNXCComp.DoStrReplace;
var
  str, strnew : string;
begin
  // StrReplace(string, idx, strnew)
  MatchString('(');
  // string
  str := GetDecoratedValue;
  CheckString;
  Next;
  MatchString(',');
  // idx
  BoolExpression;
  MatchString(',');
  // strnew
  strnew := GetDecoratedValue;
  CheckString;
  Next;
  MatchString(')');
  EmitLn(Format('strtoarr %s, %s', [StrBufName, strnew]));
  EmitLn(Format('strreplace %s, %s, %s, %s', [StrRetValName, str, RegisterName, StrBufName]));
end;

procedure TNXCComp.DoStrCat;
var
  asmStr : string;
begin
  // StrCat(str1, str2, ..., strN)
  MatchString('(');
  CheckString;
  asmStr := Format('strcat %s, %s', [StrRetValName, GetDecoratedValue]);
  Next;
  while Token = ',' do begin
    Next;
    CheckString;
    asmStr := asmStr + ', ' + GetDecoratedValue;
    Next;
  end;
  MatchString(')');
  EmitLn(asmStr);
end;

procedure TNXCComp.DoSubString;
var
  str, idx : string;
begin
  // SubStr(string, idx, len)
  MatchString('(');
  // string
  str := GetDecoratedValue;
  CheckString;
  Next;
  MatchString(',');
  // idx
  BoolExpression;
  push;
  idx := tos;
  EmitLn(Format('mov %s, %s', [idx, RegisterName]));
  MatchString(',');
  // len
  BoolExpression;
  MatchString(')');
  EmitLn(Format('strsubset %s, %s, %s, %s', [StrRetValName, str, idx, RegisterName]));
  pop;
end;

const
  APIF_ABS              = 0;
  APIF_SIGN             = 1;
  APIF_RANDOM           = 2;
  APIF_GETINPUT         = 3;
  APIF_GETOUTPUT        = 4;
  APIF_RESETSCREEN      = 7;
  APIF_TEXTOUT          = 8;
  APIF_NUMOUT           = 9;
  APIF_PLAYTONEEX       = 12;
  APIF_PLAYFILEEX       = 13;
  APIF_BUTTONPRESSED    = 18;
  APIF_BUTTONCOUNT      = 19;
  APIF_READBUTTONEX     = 20;
  APIF_DRAWPOINT        = 22;
  APIF_DRAWLINE         = 23;
  APIF_DRAWCIRCLE       = 24;
  APIF_DRAWRECT         = 25;
  APIF_DRAWGRAPHIC      = 26;
  APIF_STRTONUM         = 31;
  APIF_STRLEN           = 32;
  APIF_STRINDEX         = 33;
  APIF_ASM              = 34;

procedure TNXCComp.DoCallAPIFunc(procname: string);
var
  arg, parg, op, asmStr : string;
  id : integer;
  dt : char;
begin
  fCCSet := False;
  id := APIFuncNameToID(procname);
  case id of
    APIF_ASM : begin
      dt := #0;
      DoAsm(dt);
    end;
    APIF_RANDOM : begin
      MatchString('(');
      if Value = ')' then
      begin
        MatchString(')');
        EmitLn(Format('SignedRandom(%0:s)', [RegisterName]));
      end
      else
      begin
        BoolExpression;
        MatchString(')');
        EmitLn(Format('Random(%0:s, %0:s)', [RegisterName]));
      end;
    end;
    APIF_ABS, APIF_SIGN :
    begin
      MatchString('(');
      BoolExpression;
      MatchString(')');
      case id of
        APIF_ABS : asmStr := 'abs %0:s, %0:s';
        APIF_SIGN : asmStr := 'sign %0:s, %0:s';
      end;
      EmitLn(Format(asmStr, [RegisterName]));
    end;
    APIF_GETINPUT, APIF_GETOUTPUT :
    begin
      // GetInput(port, field)
      // GetOutput(port, field)
      MatchString('(');
      // port
      parg := GetDecoratedValue;
      Next;
      MatchString(',');
      // field
      arg := Value;
      CheckNumeric;
      Next;
      MatchString(')');
      case id of
        APIF_GETINPUT  : op := 'getin';
        APIF_GETOUTPUT : op := 'getout';
      end;
      EmitLn(Format('%s %s, %s, %s', [op, RegisterName, parg, arg]));
    end;
    APIF_RESETSCREEN : DoResetScreen;
    APIF_TEXTOUT, APIF_NUMOUT : DoTextNumOut(id);
    APIF_PLAYTONEEX : DoPlayToneEx;
    APIF_PLAYFILEEX : DoPlayFileEx;
    APIF_BUTTONPRESSED, APIF_BUTTONCOUNT, APIF_READBUTTONEX : DoReadButton(id);
    APIF_DRAWPOINT : DoDrawPoint;
    APIF_DRAWLINE, APIF_DRAWRECT : DoDrawLineRect(id);
    APIF_DRAWCIRCLE : DoDrawCircle;
    APIF_DRAWGRAPHIC : DoDrawGraphic;
    APIF_STRTONUM : DoStrToNum;
    APIF_STRLEN : DoStrLen;
    APIF_STRINDEX : DoStrIndex;
  end;
end;

procedure TNXCComp.DoStrIndex;
var
  arg : string;
begin
  // StrIndex(string, idx)
  MatchString('(');
  // string
  arg := GetDecoratedValue;
  CheckString;
  Next;
  MatchString(',');
  // idx
  BoolExpression;
  MatchString(')');
  EmitLn(Format('strindex %0:s, %s, %0:s',[RegisterName, arg]));
end;

procedure TNXCComp.DoStrLen;
var
  arg : string;
begin
  // StrLen(string)
  MatchString('(');
  // string
  arg := GetDecoratedValue;
  CheckString;
  Next;
  MatchString(')');
  EmitLn(Format('strlen %s, %s', [RegisterName, arg]));
end;

procedure TNXCComp.DoStrToNum;
var
  arg : string;
begin
  // StrToNum(string)
  MatchString('(');
  // string
  arg := GetDecoratedValue;
  CheckString;
  Next;
  MatchString(')');
  push;
  EmitLn(Format('strtonum %0:s, %s, %s, NA, NA', [RegisterName, tos, arg]));
  pop;
end;

procedure TNXCComp.DoResetScreen;
begin
  MatchString('(');
  MatchString(')');
  EmitLn('acquire __SSMArgsMutex');
  EmitLn('set __SSMArgs.ScreenMode, 0');
  EmitLn('syscall SetScreenMode, __SSMArgs');
  EmitLn(Format('mov %s, __SSMArgs.Result', [RegisterName]));
  EmitLn('release __SSMArgsMutex');
end;

procedure TNXCComp.DoTextNumOut(idx: integer);
var
  x, y, cls, txt : string;
begin
  //TextOut(x,y,cls,txt)
  //NumOut(x,y,cls,num)
  MatchString('(');
  // arg1 = x
  BoolExpression;
  push;
  x := tos;
  EmitLn(Format('mov %s, %s', [x, RegisterName]));
  MatchString(',');
  // arg2 = y
  BoolExpression;
  push;
  y := tos;
  EmitLn(Format('mov %s, %s', [y, RegisterName]));
  MatchString(',');
  // arg3 = cls
  BoolExpression;
  push;
  cls := tos;
  EmitLn(Format('mov %s, %s', [cls, RegisterName]));
  MatchString(',');
  if idx = APIF_NUMOUT then
  begin
    BoolExpression;
    MatchString(')');
    EmitLn('acquire __TextOutMutex');
    EmitLn('mov __TextOutArgs.Location.X, ' + x);
    EmitLn('mov __TextOutArgs.Location.Y, ' + y);
    EmitLn('mov __TextOutArgs.Options, ' + cls);
    EmitLn(Format('numtostr __TextOutArgs.Text, %s',[RegisterName]));
    EmitLn('syscall DrawText, __TextOutArgs');
    EmitLn(Format('mov %s, __TextOutArgs.Result',[RegisterName]));
    EmitLn('release __TextOutMutex');
  end
  else
  begin
    txt := GetDecoratedValue;
    CheckString;
    Next;
    MatchString(')');
    EmitLn('acquire __TextOutMutex');
    EmitLn('mov __TextOutArgs.Location.X, ' + x);
    EmitLn('mov __TextOutArgs.Location.Y, ' + y);
    EmitLn('mov __TextOutArgs.Options, ' + cls);
    EmitLn('mov __TextOutArgs.Text, ' + txt);
    EmitLn('syscall DrawText, __TextOutArgs');
    EmitLn(Format('mov %s, __TextOutArgs.Result',[RegisterName]));
    EmitLn('release __TextOutMutex');
  end;
  pop;
  pop;
  pop;
end;

procedure TNXCComp.DoDrawPoint;
var
  x, y : string;
begin
  //DrawPoint(x,y,cls)
  MatchString('(');
  // arg1 = x
  BoolExpression;
  push;
  x := tos;
  EmitLn(Format('mov %s, %s', [x, RegisterName]));
  MatchString(',');
  // arg2 = y
  BoolExpression;
  push;
  y := tos;
  EmitLn(Format('mov %s, %s', [y, RegisterName]));
  MatchString(',');
  // arg3 = cls
  BoolExpression;
  MatchString(')');
  EmitLn('acquire __PointOutMutex');
  EmitLn('mov __PointOutArgs.Location.X, ' + x);
  EmitLn('mov __PointOutArgs.Location.Y, ' + y);
  EmitLn(Format('mov __PointOutArgs.Options, %s', [RegisterName]));
  EmitLn('syscall DrawPoint, __PointOutArgs');
  EmitLn(Format('mov %s, __PointOutArgs.Result',[RegisterName]));
  EmitLn('release __PointOutMutex');
  pop;
  pop;
end;

procedure TNXCComp.DoDrawLineRect(idx : integer);
var
  x, y, x2, y2 : string;
begin
  //DrawLine(x1,y1,x2,y2,cls)
  //DrawRect(x1,y1,width,height,cls)
  MatchString('(');
  // arg1 = x
  BoolExpression;
  push;
  x := tos;
  EmitLn(Format('mov %s, %s', [x, RegisterName]));
  MatchString(',');
  // arg2 = y
  BoolExpression;
  push;
  y := tos;
  EmitLn(Format('mov %s, %s', [y, RegisterName]));
  MatchString(',');
  // arg3 = x2
  BoolExpression;
  push;
  x2 := tos;
  EmitLn(Format('mov %s, %s', [x2, RegisterName]));
  MatchString(',');
  // arg4 = y2
  BoolExpression;
  push;
  y2 := tos;
  EmitLn(Format('mov %s, %s', [y2, RegisterName]));
  MatchString(',');
  // arg5 = cls
  BoolExpression;
  MatchString(')');
  if idx = APIF_DRAWRECT then
  begin
    EmitLn('acquire __RectOutMutex');
    EmitLn('mov __RectOutArgs.Location.X, ' + x);
    EmitLn('mov __RectOutArgs.Location.Y, ' + y);
    EmitLn('mov __RectOutArgs.Size.Width, ' + x2);
    EmitLn('mov __RectOutArgs.Size.Height, ' + y2);
    EmitLn(Format('mov __RectOutArgs.Options, %s', [RegisterName]));
    EmitLn('syscall DrawRect, __RectOutArgs');
    EmitLn(Format('mov %s, __RectOutArgs.Result',[RegisterName]));
    EmitLn('release __RectOutMutex');
  end
  else
  begin
    EmitLn('acquire __LineOutMutex');
    EmitLn('mov __LineOutArgs.StartLoc.X, ' + x);
    EmitLn('mov __LineOutArgs.StartLoc.Y, ' + y);
    EmitLn('mov __LineOutArgs.EndLoc.X, ' + x2);
    EmitLn('mov __LineOutArgs.EndLoc.Y, ' + y2);
    EmitLn(Format('mov __LineOutArgs.Options, %s', [RegisterName]));
    EmitLn('syscall DrawLine, __LineOutArgs');
    EmitLn(Format('mov %s, __LineOutArgs.Result',[RegisterName]));
    EmitLn('release __LineOutMutex');
  end;
  pop;
  pop;
  pop;
  pop;
end;

procedure TNXCComp.DoDrawCircle;
var
  x, y, radius : string;
begin
  //DrawCircle(x1,y1,radius,cls)
  MatchString('(');
  // arg1 = x
  BoolExpression;
  push;
  x := tos;
  EmitLn(Format('mov %s, %s', [x, RegisterName]));
  MatchString(',');
  // arg2 = y
  BoolExpression;
  push;
  y := tos;
  EmitLn(Format('mov %s, %s', [y, RegisterName]));
  MatchString(',');
  // arg3 = radius
  BoolExpression;
  push;
  radius := tos;
  EmitLn(Format('mov %s, %s', [radius, RegisterName]));
  MatchString(',');
  // arg4 = cls
  BoolExpression;
  MatchString(')');
  EmitLn('acquire __CircleOutMutex');
  EmitLn('mov __CircleOutArgs.Center.X, ' + x);
  EmitLn('mov __CircleOutArgs.Center.Y, ' + y);
  EmitLn('mov __CircleOutArgs.Size, ' + radius);
  EmitLn(Format('mov __CircleOutArgs.Options, %s', [RegisterName]));
  EmitLn('syscall DrawCircle, __CircleOutArgs');
  EmitLn(Format('mov %s, __CircleOutArgs.Result',[RegisterName]));
  EmitLn('release __CircleOutMutex');
  pop;
  pop;
  pop;
end;

procedure TNXCComp.DoDrawGraphic;
var
  x, y, fname : string;
begin
  //DrawGraphic(x,y,fname,cls)
  MatchString('(');
  // arg1 = x
  BoolExpression;
  push;
  x := tos;
  EmitLn(Format('mov %s, %s', [x, RegisterName]));
  MatchString(',');
  // arg2 = y
  BoolExpression;
  push;
  y := tos;
  EmitLn(Format('mov %s, %s', [y, RegisterName]));
  MatchString(',');
  // arg3 = fname
  fname := GetDecoratedValue;
  CheckString;
  Next;
  MatchString(',');
{
  // arg4 = vars
  vars := GetDecoratedValue;
  Next;
  MatchString(',');
}
  // arg4 = cls
  BoolExpression;
  MatchString(')');
  EmitLn('acquire __GraphicOutMutex');
  EmitLn('mov __GraphicOutArgs.Location.X, ' + x);
  EmitLn('mov __GraphicOutArgs.Location.Y, ' + y);
  EmitLn('mov __GraphicOutArgs.Filename, ' + fname);
//  EmitLn('mov __GraphicOutArgs.Variables, ' + vars);
  EmitLn(Format('mov __GraphicOutArgs.Options, %s', [RegisterName]));
  EmitLn('syscall DrawGraphic, __GraphicOutArgs');
  EmitLn(Format('mov %s, __GraphicOutArgs.Result',[RegisterName]));
  EmitLn('release __GraphicOutMutex');
  pop;
  pop;
end;

procedure TNXCComp.DoPlayToneEx;
var
  freq, dur, vol : string;
begin
  //PlayToneEx(freq, dur, vol, loop)
  MatchString('(');
  // arg1 == Frequency
  BoolExpression;
  push;
  freq := tos;
  EmitLn(Format('mov %s, %s', [freq, RegisterName]));
  MatchString(',');
  // arg2 == Duration
  BoolExpression;
  push;
  dur := tos;
  EmitLn(Format('mov %s, %s', [dur, RegisterName]));
  MatchString(',');
  // arg3 == Volume
  BoolExpression;
  push;
  vol := tos;
  EmitLn(Format('mov %s, %s', [vol, RegisterName]));
  MatchString(',');
  // arg4 == loop?
  BoolExpression;
  MatchString(')');
  EmitLn('acquire __SPTArgsMutex');
  EmitLn('mov __SPTArgs.Frequency, ' + freq);
  EmitLn('mov __SPTArgs.Duration, ' + dur);
  EmitLn('mov __SPTArgs.Volume, ' + vol);
  EmitLn(Format('mov __SPTArgs.Loop, %s',[RegisterName]));
  EmitLn('syscall SoundPlayTone, __SPTArgs');
  EmitLn(Format('mov %s, __SPTArgs.Result',[RegisterName]));
  EmitLn('release __SPTArgsMutex');
  pop;
  pop;
  pop;
end;

procedure TNXCComp.DoPlayFileEx;
var
  fname, vol : string;
begin
  //PlayFileEx(file, vol, loop?)
  MatchString('(');
  // arg1 == Filename (either constant string or string variable)
  fname := GetDecoratedValue;
  CheckString;
  Next;
  MatchString(',');
  // arg2 == Volume
  BoolExpression;
  push;
  vol := tos;
  EmitLn(Format('mov %s, %s', [vol, RegisterName]));
  MatchString(',');
  // arg3 == loop?
  BoolExpression;
  MatchString(')');
  EmitLn('acquire __SPFArgsMutex');
  EmitLn('mov __SPFArgs.Filename, ' + fname);
  EmitLn('mov __SPFArgs.Volume, ' + vol);
  EmitLn(Format('mov __SPFArgs.Loop, %s', [RegisterName]));
  EmitLn('syscall SoundPlayFile, __SPFArgs');
  EmitLn(Format('mov %s, __SPFArgs.Result', [RegisterName]));
  EmitLn('release __SPFArgsMutex');
  pop;
end;

procedure TNXCComp.DoReadButton(idx: integer);
var
  btn, pressed, count : string;
begin
  // ButtonPressed(btn, reset)
  // ButtonCount(btn, reset)
  // ReadButtonEx(btn, reset, pressed, count)
  MatchString('(');
  // arg1 = button index
  BoolExpression;
  push;
  btn := tos;
  EmitLn(Format('mov %s, %s', [btn, RegisterName]));
  MatchString(',');
  // arg2 = reset?
  BoolExpression;
  if idx = APIF_READBUTTONEX then
  begin
    // two output args
    MatchString(',');
    // pressed
    pressed := GetDecoratedValue;
    CheckIdent;
    CheckTable(Value);
    Next;
    MatchString(',');
    // count
    count := GetDecoratedValue;
    CheckIdent;
    CheckTable(Value);
    Next;
  end;
  MatchString(')');
  EmitLn('acquire __RBtnMutex');
  EmitLn('mov __RBtnArgs.Index, ' + btn);
  EmitLn(Format('mov __RBtnArgs.Reset, %s', [RegisterName]));
  EmitLn('syscall ReadButton, __RBtnArgs');
  if idx = APIF_BUTTONCOUNT then
    EmitLn(Format('mov %s, __RBtnArgs.Count',[RegisterName]))
  else if idx = APIF_BUTTONPRESSED then
    EmitLn(Format('mov %s, __RBtnArgs.Pressed',[RegisterName]))
  else
  begin
    EmitLn(Format('mov %s, __RBtnArgs.Pressed', [pressed]));
    EmitLn(Format('mov %s, __RBtnArgs.Count', [count]));
    EmitLn(Format('mov %s, __RBtnArgs.Result', [RegisterName]));
  end;
  EmitLn('release __RBtnMutex');
  pop;
end;

procedure TNXCComp.LoadAPIFunctions;
begin
  AddAPIFunction('asm', APIF_ASM);
  AddAPIFunction('abs', APIF_ABS);
  AddAPIFunction('sign', APIF_SIGN);
  AddAPIFunction('Random', APIF_RANDOM);
  AddAPIFunction('GetInput', APIF_GETINPUT);
  AddAPIFunction('GetOutput', APIF_GETOUTPUT);
  AddAPIFunction('ResetScreen', APIF_RESETSCREEN);
  AddAPIFunction('TextOut', APIF_TEXTOUT);
  AddAPIFunction('NumOut', APIF_NUMOUT);
  AddAPIFunction('PlayToneEx', APIF_PLAYTONEEX);
  AddAPIFunction('PlayFileEx', APIF_PLAYFILEEX);
  AddAPIFunction('ButtonPressed', APIF_BUTTONPRESSED);
  AddAPIFunction('ButtonCount', APIF_BUTTONCOUNT);
  AddAPIFunction('ReadButtonEx', APIF_READBUTTONEX);
  AddAPIFunction('PointOut', APIF_DRAWPOINT);
  AddAPIFunction('LineOut', APIF_DRAWLINE);
  AddAPIFunction('CircleOut', APIF_DRAWCIRCLE);
  AddAPIFunction('RectOut', APIF_DRAWRECT);
  AddAPIFunction('GraphicOut', APIF_DRAWGRAPHIC);
  AddAPIFunction('StrToNum', APIF_STRTONUM);
  AddAPIFunction('StrLen', APIF_STRLEN);
  AddAPIFunction('StrIndex', APIF_STRINDEX);
  AddAPIStringFunction('NumToStr', APISF_NUMTOSTR);
  AddAPIStringFunction('StrCat', APISF_STRCAT);
  AddAPIStringFunction('SubStr', APISF_SUBSTR);
  AddAPIStringFunction('Flatten', APISF_FLATTEN);
  AddAPIStringFunction('StrReplace', APISF_STRREPLACE);
end;

(*
FileOpenRead (0), FileOpenWrite (1), FileOpenAppend (2), FileRead (3),
FileWrite (4), FileClose (5), FileResolveHandle (6), FileRename (7),
FileDelete (8),
CommLSWrite (21), CommLSRead (22), CommLSCheckStatus (23),
MessageWrite (26), MessageRead (27),
CommBTCheckStatus (28), CommBTWrite (29), IOMapRead (32),
IOMapWrite (33)
*)

end.
